import copy
import time
from typing import Optional, List, Dict, Union

import numpy as np
from scipy.special import binom 
from tqdm import tqdm

from qrem.functions_qrem import ancillary_functions as anf
from qrem.noise_characterization.base_classes.overlapping_estimation_base import OverlappingEstimationBase
from qrem.noise_characterization.tomography_design.overlapping import overlapping_tomography_functions as otf

from qrem.common import convert, math
from qrem.common.printer import qprint

class OverlappingTomographyBase(OverlappingEstimationBase):
    """


    """
    def __init__(self,
                 number_of_qubits: int,
                 experiment_name: str,
              #   maximal_circuits_number: Optional[int] = 1500,  #This can be retreived latter
                 subsets_list: Optional[List[List[int]]] = None, #for consitency I would use subsets for this variable or subset_list -conflict with parent class terminology
                 subsets_locality: Optional[int] = 2,
                 show_progress_bars: Optional[bool] = False
                 ):

        """

        :param number_of_qubits:
        :param number_of_symbols: number of different experimental setups to consider
                                 for example - Diagonal Detector Overlapping Tomography (DDOT)
                                 requires only 2 symbols corresponding to X and identity gates

        :param subsets_list: list of subsets of qubits for which marginal tomography
                             should be performed
                             NOTE: for tomography which is to be performed NOT on ALL subsets
                             of given locality, this must be provided

        :param subsets_locality: locality of ALL considered subsets
                                 NOTE: this should be provided only if ALL subsets have the same
                                       locality


        #TODO: describe rough functionalities and ligh level logic of this class (depending on optional parameters give n)
        """
        # when no subsets are given and no locality of subsets provided return erros 
        if subsets_list is None and subsets_locality is None:
            raise ValueError('Please provide subsets list or desired locality')
        #  when subset_list is empty but subset_locality is given - generate all k element
        elif subsets_list is None and subsets_locality is not None:
            subsets_list = math.get_k_local_subsets(number_of_elements=number_of_qubits, 
                                                         subset_size=subsets_locality)


        # depending on type of experiments use different number of symbols for circuits      
        #ORGANIZATION - nie do konca rozumiem czym QDT rozni sie od QDOT - (pp) z tego co rozuzmiem niczym, po prostu możn użzyć obu nazw
        if experiment_name.upper() == 'QDT' or experiment_name.upper() == 'QDOT':
            number_of_symbols = 6
        elif experiment_name.upper() == 'DDOT':
            number_of_symbols=2

# ORGANIZATION : check logic of the incheritance of this class (MO)
        super().__init__(number_of_qubits=number_of_qubits,
                         number_of_symbols=number_of_symbols,
                         subsets=subsets_list,
                  #       maximal_circuits_amount=maximal_circuits_number, #this can be retreived latter
                         show_progress_bars=show_progress_bars

                         )

#asignement of subset locality variable to local variant 
        self._subsets_locality = subsets_locality
        if subsets_locality is not None:


#ORGANIZATION: variable _number_of_elements_collection is an upper bound on the number of elements needed for "perfect tomography"
# This variable is used exclusivelly in internal function _compute_perfect_collection_bruteforce. Hence this is a candidate to removal (MO)

            self._elements_in_perfect_collection = int(
                binom(self._number_of_qubits,self._subsets_locality) *
                self._number_of_symbols ** self._subsets_locality)
        else:
            self._elements_in_perfect_collection = int(sum(
                [number_of_symbols ** len(subset_now) for subset_now in subsets_list]))

#------------------------------

        '''
        I changed subset_list to subset for consistency in the comprehension below (MO)
        Dictionary below has keys that label qubit tuples (in the format q0q1 etc. ), for each key the value is an array
        storing occurrences of arrangements of number_of_symbols among  |size if subset| slots (there are number_of_symbols ** len(subset)) of them
        This structure is presumably used to test the existence of families of perfect functions

        dtape=int is there to indicate integer occurrences 

        '''

        #MOcomm - is it the best thing to do alghorimically? perhaps other data structures are faster? 
        # For example -kays can be just given by subsets and (perhaps) it would be faster to iterate over them

        #deleted redundant "not private" version of this dictionary
        self._dictionary_symbols_counting = {
            anf.get_qubits_keystring(subset): np.zeros(self._number_of_symbols ** len(subset),
                                                       dtype=int) for 
            subset in self._subsets}


#MOcomm I constructed properties dictionary in one line 
#MOcomm - not sure what is the meaning of this dictionary and these parameters
        self._circuits_properties_dictionary = {'number_of_circuits': 0,
                                          'absent_elements_amount': 10 ** 6,
                                          'minimal_amount': 10 ** 6,
                                          'maximal_amount': 10 ** 6,
                                          'median': 10 ** 6,
                                          'mean': 10 ** 6,
                                          'SD': 10 ** 6}


        #MOcomm - not sure what is the meaning of this
        if number_of_symbols < 11:
            self._integer_representation = lambda integer: str(integer)
        else:
            self._integer_representation = lambda integer: self.integer_to_charstring(integer)

    @property
    def dictionary_symbols_counting(self) -> Dict[str, np.ndarray]:
        return self._dictionary_symbols_counting

    @property
    def circuits_properties_dictionary(self) -> Dict[str, float]:
        return self._circuits_properties_dictionary

    @property
    def circuits_list(self) -> List[List[int]]:
        return self._circuits_list

    @circuits_list.setter
    def circuits_list(self, circuits_list: List[List[int]]):
        self.reset_everything()
        self._circuits_list = circuits_list


#ORGANIZE- in my opinion there is no need to have aditional dictionary_symbols_counting dicrtionary
#We can add it to begin with to private dict (MO)

    def add_dictionary_subsets_symbols_counting_template(self):
        dictionary_symbols_counting = {
            anf.get_qubits_keystring(subset_list): np.zeros(self._number_of_symbols ** len(subset_list),
                                                       dtype=int) for
            subset_list in self._subsets}

        self._dictionary_symbols_counting = dictionary_symbols_counting

    #MOVE_TO >> core.utils 
    def integer_to_charstring(integer: int):
        if integer < 10:
            return str(integer)
        else:
            return chr(integer + 97 - 10)

# MOcomm - not clear the meaning of it
    def update_dictionary_subsets_symbols_counting(self,
                                                   circuits_list: Optional[List[List[Union[int,str]]]] = None,
                                                   count_added_subcircuits: Optional[bool] = False):
        if circuits_list is None:
            circuits_list = self._circuits_list

        subsets_range = range(len(self._subsets))
        if self._show_progress_bars:
            subsets_range = tqdm(subsets_range)

        if count_added_subcircuits:
            added_subcircuits_counter = 0
        else:
            added_subcircuits_counter = None

        for subset_index in subsets_range:
            for circuit in circuits_list:
                qubits_key = anf.get_qubits_keystring(self._subsets[subset_index])

                subset_symbols_now = [circuit[qubit_index] for qubit_index in self._subsets[subset_index]]

                # print(subset_symbols_now, any(s > self._number_of_symbols for s in subset_symbols_now))

                if any(s >= self._number_of_symbols for s in subset_symbols_now):
                    pass
                else:
                    subset_circuit_identifier = int(
                        ''.join([self._integer_representation(s) for s in subset_symbols_now]),
                        self._number_of_symbols)

                    if count_added_subcircuits:
                        if self._dictionary_symbols_counting[qubits_key][subset_circuit_identifier] == 0:
                            added_subcircuits_counter += 1

                    self._dictionary_symbols_counting[qubits_key][subset_circuit_identifier] += 1

        if count_added_subcircuits:
            return added_subcircuits_counter

    def get_absent_symbols_amount(self):

        t0 = time.time()
        zero_subsets = 0
        for subset in self._subsets:
            subset_counts = self._dictionary_symbols_counting[anf.get_qubits_keystring(subset)]
            zero_subsets += len(subset_counts) - np.count_nonzero(subset_counts)

        self._circuits_properties_dictionary['absent_elements_amount'] = zero_subsets
        qprint('This took:', time.time() - t0)

        return zero_subsets
    #
    def get_absent_symbol_indices(self):
        zero_subsets = {}
        for subset in tqdm(self._subsets):
            subset_counts = self._dictionary_symbols_counting[anf.get_qubits_keystring(subset)]
            stuff_now = []
            for index in range(len(subset_counts)):
                if subset_counts[index]==0:
                    stuff_now.append(index)
            zero_subsets[subset] = stuff_now

        return zero_subsets

    def calculate_properties_of_circuits(self) -> None:

        big_list = []
        for subset in self._subsets:
            big_list += list(self._dictionary_symbols_counting[anf.get_qubits_keystring(subset)])

        minimal_amount, maximal_amount = min(big_list), max(big_list)

        big_list = np.array(big_list)

        mean, SD, median = np.mean(big_list), np.std(big_list), np.median(big_list)

        absent_elements_amount = len(big_list) - np.count_nonzero(big_list)

        self._circuits_properties_dictionary['number_of_circuits'] = len(self._circuits_list)
        self._circuits_properties_dictionary['absent_elements_amount'] = absent_elements_amount
        self._circuits_properties_dictionary['minimal_amount'] = minimal_amount
        self._circuits_properties_dictionary['maximal_amount'] = maximal_amount
        self._circuits_properties_dictionary['median'] = median
        self._circuits_properties_dictionary['mean'] = mean
        self._circuits_properties_dictionary['SD'] = SD


    #MOcomm a function reseting everything in an object
    def reset_everything(self):
        #MOcomment - I constructed private dictionary with properties in one line now
        self._circuits_properties_dictionary= {'number_of_circuits': 0,
                                          'absent_elements_amount': 10 ** 6,
                                          'minimal_amount': 10 ** 6,
                                          'maximal_amount': 10 ** 6,
                                          'median': 10 ** 6,
                                          'mean': 10 ** 6,
                                          'SD': 10 ** 6}

        self._circuits_list = []

#changed argument in this comprehension
        dictionary_symbols_counting = {
            anf.get_qubits_keystring(subset): np.zeros(self._number_of_symbols ** len(subset),
                                                       dtype=int) for
            subset in self._subsets}

        self._dictionary_symbols_counting = dictionary_symbols_counting

    def _cost_function_circuits_amount(self):
        return len(self._circuits_list)

    def _cost_function_circuits_SD(self):
        return self._circuits_properties_dictionary['SD']

    def _cost_function_minimal_amount_of_circuits(self):
        return -self._circuits_properties_dictionary['minimal_amount']

    def _cost_function_absent_elements(self):
        return self._circuits_properties_dictionary['absent_elements_amount']

    def _cost_function_maximal_spread(self):
        return self._circuits_properties_dictionary['maximal_amount'] - \
               self._circuits_properties_dictionary['minimal_amount']




    def __compute_1q_collection(self):

        circuits_now = [[unique_symbol for _ in range(self._number_of_qubits)] for unique_symbol in range(self._number_of_symbols)]
        self.add_circuits(circuits_now)
        self.update_dictionary_subsets_symbols_counting(
            circuits_list=circuits_now,
            count_added_subcircuits=True)

# ORGANIZE - after discussion with Janek we decided that this function should be updated,
#  now it is highly nonoptimal, if not wrong (MO)
    def _compute_perfect_collection_bruteforce(self,
                                               circuits_in_batch: int,
                                               print_updates: bool
                                               ):

        runs_number = 1
        absent_elements_amount = self._elements_in_perfect_collection

        while absent_elements_amount > 0 and len(self.circuits_list) < self._maximal_circuits_amount:
            if runs_number % 20 == 0 and print_updates:
                qprint('Run number:', runs_number)
                qprint('Number of circuits:', len(self._circuits_list))
                qprint('Absent elements amount:', absent_elements_amount)

            circuits_now = self.get_random_circuits_list(circuits_in_batch)
            self.add_circuits(circuits_now)

            added_elements = self.update_dictionary_subsets_symbols_counting(
                circuits_list=circuits_now,
                count_added_subcircuits=True)
            absent_elements_amount -= added_elements

            runs_number += 1




    def _get_proper_cost_function(self,
                                  optimized_quantity: str):

        if optimized_quantity in ['circuits_amount', 'circuits_number', 'amount', 'circuits']:
            cost_function = self._cost_function_circuits_amount
        elif optimized_quantity in ['std', 'SD', 'standard_deviation']:
            cost_function = self._cost_function_circuits_SD
        elif optimized_quantity in ['minimal_amount']:
            cost_function = self._cost_function_minimal_amount_of_circuits
        elif optimized_quantity in ['spread', 'maximal_spread']:
            cost_function = self._cost_function_maximal_spread
        elif optimized_quantity in ['absent', 'absent_elements']:
            cost_function = self._cost_function_absent_elements
        else:
            raise ValueError('Wrong optimized quantity string: ' + optimized_quantity + '.')

        return cost_function

    def _add_cost_functions(self,
                            dictionary_cost_functions: Dict[str, float]):

        def cost_functions_added():
            returned_quantity = 0
            for function_label, function_weight in dictionary_cost_functions.items():
                returned_quantity += function_weight * self._get_proper_cost_function(function_label)()
            return returned_quantity

        return cost_functions_added

    def _compute_perfect_collection_bruteforce_randomized(self,
                                                          number_of_iterations: int,
                                                          circuits_in_batch: int,
                                                          print_updates: bool,
                                                          optimized_quantity: Union[
                                                              str, Dict[str, float]],
                                                          additional_circuits: Optional[int] = 0,
                                                          fixed_circuits:Optional[List[List[int]]]=None
                                                          ):
        """
        This function implements self._compute_perfect_collection_bruteforce
        for number_of_iterations times, then adds additional_circuits number of random circuits,
        computes cost function and chooses the family that minimizes cost function.

        :param number_of_iterations: how many times random perfect family should be generated
        :param circuits_in_batch: see self._compute_perfect_collection_bruteforce
        :param print_updates: whether to print updates during optimization
        :param optimized_quantity: specify what cost function is
        Possible string values:
        1. 'minimal_amount' - maximizes number of least-frequent subset-circuits
        2. 'spread' - minimizes difference between maximal and minimal number of subset-circuits

        3. 'circuits_amount' - minimizes number of circuits
                        (NOTE: it does not make sense to choose this option with additional_circuits>0)
        4. 'SD' - minimizes standard deviation of occurrences of subset-circuits

        It is possible to use combined cost functions_qrem.
        Dictionary must be provided where KEY is label for optimized quantity and VALUE is its weight.

        For example:
        optimized_quantity = {'minimal_amount': 1.0,
                              'spread':0.5}

        will give cost function which returns 1.0 * (-number of least frequent circuits)
                                          + 0.5 * (difference between most and least fequenet circuits)


        :param additional_circuits: number of circuits which are to be added to the PERFECT collection
                                    obtained in optimization loop. Those are "additional" circuits in
                                    a sense that they are not needed for collection to be perfect,
                                    but instead are used to obtain better values of cost function
                                    or just add more experiments reduce statistical noise.
        :return:
        """

        if isinstance(optimized_quantity, str):
            cost_function = self._get_proper_cost_function(optimized_quantity=optimized_quantity)
        elif isinstance(optimized_quantity, dict):
            cost_function = self._add_cost_functions(dictionary_cost_functions=optimized_quantity)
        else:
            raise ValueError("No cost function was specified!")


        runs_range = range(number_of_iterations)
        if self._show_progress_bars:
            runs_range = tqdm(runs_range)

        # circuit_families = []
        # best_family = None
        global_cost, best_family = 10 ** 6, None

        for runs_number in runs_range:
            if runs_number % int(np.ceil(number_of_iterations / 20)) == 0 and print_updates:
                qprint('Run number:', runs_number, 'red')
                qprint('Current best value of cost function:', global_cost)

            self.reset_everything()

            if fixed_circuits is not None:
                self.add_circuits(fixed_circuits)
                self.update_dictionary_subsets_symbols_counting()



            self._compute_perfect_collection_bruteforce(circuits_in_batch=circuits_in_batch,
                                                        print_updates=False
                                                        )

            if additional_circuits > 0:
                current_length, maximal_length = len(
                    self._circuits_list), self._maximal_circuits_amount
                if current_length < maximal_length:
                    if additional_circuits > maximal_length - current_length:
                        adding_circuits = maximal_length - current_length
                    else:
                        adding_circuits = copy.deepcopy(additional_circuits)
                    # print(adding_circuits)
                    new_circuits = self.get_random_circuits_list(adding_circuits)

                    self.update_dictionary_subsets_symbols_counting(new_circuits)
                    self.add_circuits(new_circuits)

            self.calculate_properties_of_circuits()

            cost_now = cost_function()

            if cost_now < global_cost:
                best_family = copy.deepcopy(self._circuits_list)
                global_cost = cost_now

        qprint('best family length', len(best_family), 'red')
        self.reset_everything()
        self._circuits_list = best_family
        self.update_dictionary_subsets_symbols_counting()
        self.calculate_properties_of_circuits()

    def _add_absent_elements_by_hand(self):

        best_family = copy.deepcopy(self._circuits_list)

        missing_elements = self.get_absent_symbol_indices()

        number_of_added_circuits_multiplier = 3

        added_circuits = []
        for subset, circuit_indices_list in tqdm(missing_elements.items()):
            register_local = dict(enumerate(convert.get_ditstrings_register(base=self._number_of_symbols,
                                                         number_of_dits=len(subset))))

            for _ in range(number_of_added_circuits_multiplier):
                for circuit_index in circuit_indices_list:
                    circuit_now_global = self.get_random_circuit()

                    local_circuit_now = register_local[circuit_index]

                    for symbol_index in range(len(subset)):
                        circuit_now_global[subset[symbol_index]] = local_circuit_now[symbol_index]
                    added_circuits.append(circuit_now_global)


        # self.reset_everything()
        # print(len(added_circuits))
        self._show_progress_bars = True
        self.update_dictionary_subsets_symbols_counting(circuits_list=added_circuits)
        self._circuits_list = best_family + added_circuits
        self.calculate_properties_of_circuits()

    #(PP) same as in base class?
    def compute_random_collection(self,
                                  number_of_circuits=None):

        if number_of_circuits is None:
            #TODO FBM: add probabilistic bounds
            raise ValueError()


        self._circuits_list = self.add_random_circuits(number_of_circuits)

        return self._circuits_list


    # (PP) not used as of now, commenting 
    # (PP) TODO_MO - what is the usecase for this function?
    # def optimize_perfect_collection(self,
    #                                 method_name='bruteforce_randomized',
    #                                 method_kwargs=None):

    #     """
    #     Find perfect collection of overlapping circuits.
    #     "Perfect" means that for each subset of qubits self._subsets_list[i],
    #      each symbol out of self._number_of_symbols^self._subsets_list[i],
    #      appears in the collection at least once.

    #     :param method_name:
    #     possible values:

    #     1. 'bruteforce' - see self._compute_perfect_collection_bruteforce
    #     2. 'bruteforce_randomized' - see self._compute_perfect_collection_bruteforce_randomized

    #     :param method_kwargs: kwargs for chosen method, see corresponding methods' descriptions
    #     :return:
    #     """

    #     if self._subsets_locality == 1:
    #         self.__compute_1q_collection()

    #     else:

    #         if method_name == 'bruteforce':
    #             if method_kwargs is None:
    #                 method_kwargs = {'circuits_in_batch': 1,
    #                                  'print_updates': True}

    #             self._compute_perfect_collection_bruteforce(**method_kwargs)

    #         elif method_name == 'bruteforce_randomized':
    #             if method_kwargs is None:
    #                 method_kwargs = {'number_of_iterations': 100,
    #                                  'circuits_in_batch': 1,
    #                                  'print_updates': True,
    #                                  'optimized_quantity': 'minimal_amount',
    #                                  'additional_circuits': 0,
    #                                  'fixed_circuits':None}

    #             self._compute_perfect_collection_bruteforce_randomized(**method_kwargs)


    #         absent_elements = self._circuits_properties_dictionary['absent_elements_amount']

    #         if absent_elements != 0:
    #             qprint('________WARNING________:',
    #                            'The collection is not perfect. '
    #                            'It is missing %s' % absent_elements
    #                            + ' elements!\nTry increasing limit on the circuits amount.',
    #                            'red')

    #             if anf.query_yes_no("Do you wish to add missing elements by hand?"):
    #                 self._add_absent_elements_by_hand()

    #         absent_elements = self._circuits_properties_dictionary['absent_elements_amount']

    #         if absent_elements != 0:
    #             qprint('________WARNING________:',
    #                            'The collection is not perfect. '
    #                            'It is missing %s' % absent_elements
    #                            + ' elements!\nTry increasing limit on the circuits amount.',
    #                            'red')

    #             if anf.query_yes_no("Do you wish to add missing elements by hand?"):
    #                 self._add_absent_elements_by_hand()

    def get_parallel_tomography_on_subsets(self,
                                           number_of_circuits: int,
                                           non_overlapping_subsets:List[List[int]]=None,
                                           ) -> List[List[int]]:
        """

        :param number_of_circuits: should be power of 2
        :param subsets:
        :param number_of_symbols:
        :return:
        :rtype:
        """

        if non_overlapping_subsets is None:
            non_overlapping_subsets = self._subsets

        number_of_symbols = self._number_of_symbols

        return otf.get_parallel_tomography_on_non_overlapping_subsets(
            number_of_circuits=number_of_circuits,
            number_of_symbols=number_of_symbols,
            non_overlapping_subsets=non_overlapping_subsets)



# -*- coding: utf-8 -*-
"""
Created on Tue Jan 27 11:41:52 2015

@author: wittek
"""
from __future__ import division, print_function
import glob
from math import copysign
import numpy as np
import os
from scipy.sparse import lil_matrix
from .sdp_relaxation import Relaxation


def collinsgisin_to_faacets(I):
    coefficients = []
    for j in range(len(I[0])):
        for i in range(len(I)):
            coefficients.append(I[i][j])
    return np.array(coefficients, dtype='d')


def configuration_to_faacets(A_configuration, B_configuration):
    a = str(A_configuration).replace('[', '(').replace(']', ')').replace(',',
                                                                         '')
    b = str(B_configuration).replace('[', '(').replace(']', ')').replace(',',
                                                                         '')
    return '[' + a + ' ' + b + ']'


def get_faacets_moment_matrix(A_configuration, B_configuration, coefficients):
    from jpype import startJVM, shutdownJVM, JPackage, getDefaultJVMPath, \
        JArray, JDouble
    # find the JAR file and start the JVM
    jarFiles = glob.glob('faacets-*.jar')
    assert len(jarFiles) == 1
    jarFile = os.path.join(os.getcwd(), jarFiles[0])
    startJVM(getDefaultJVMPath(), "-Djava.class.path="+jarFile)
    com = JPackage('com')

    # main code
    sc = com.faacets.Core.Scenario(configuration_to_faacets(A_configuration,
                                                            B_configuration))
    representation = com.faacets.Core.Representation('NCRepresentation')
    ope = com.faacets.SDP.OperatorElements(['', 'A', 'AB'])
    pts = com.faacets.SDP.PartialTransposes([])
    vec = com.faacets.Core.QVector(JArray(JDouble)(coefficients))
    expr = com.faacets.Core.Expr(sc, representation, vec)
    sdp = com.faacets.SDP.CorrSDP(sc, ope, pts, expr.symmetryGroup())
    M = np.array(sdp.indexArray())
    ncIndices = np.array(sdp.ncIndices())
    shutdownJVM()
    return M, ncIndices


class FaacetsRelaxation(Relaxation):

    """Class for wrapping around a Faacets relaxation.

    """

    def get_relaxation(self, A_configuration, B_configuration, I):
        """Get the sparse SDP relaxation of a Bell inequality.

        :param A_configuration: The definition of measurements of Alice.
        :type A_configuration: list of list of int.
        :param B_configuration: The definition of measurements of Bob.
        :type B_configuration: list of list of int.
        :param I: The matrix describing the Bell inequality in the
                  Collins-Gisin picture.
        :type I: list of list of int.
        """
        coefficients = collinsgisin_to_faacets(I)
        M, ncIndices = get_faacets_moment_matrix(A_configuration,
                                                 B_configuration, coefficients)
        self.n_vars = M.max() - 1
        bs = len(M)  # The block size
        self.block_struct = [bs]
        self.F = lil_matrix((bs**2, self.n_vars + 1))
        # Constructing the internal representation of the constraint matrices
        # See Section 2.1 in the SDPA manual and also Yalmip's internal
        # representation
        for i in range(bs):
            for j in range(i, bs):
                if M[i, j] != 0:
                    self.F[i*bs+j, abs(M[i, j])-1] = copysign(1, M[i, j])
        self.obj_facvar = [0 for _ in range(self.n_vars)]
        for i in range(1, len(ncIndices)):
            self.obj_facvar[abs(ncIndices[i])-2] += \
                copysign(1, ncIndices[i])*coefficients[i]



from qrem.types.datastructure_base import DataStructureBase
from qrem.common import math
from typing import Dict, Tuple, List
import operator
from functools import reduce
import numpy as np
import itertools


from qrem.common import math, probability




#TODO: TUZJAN, prepare full local test
#TODO: Find a way of exporting to jason dictionary, in which key is a tuple. This is needed to export CNModelData to json, e.g. noise matrices are stored in a dictionary with tupels as keys.  
class CNModelData(DataStructureBase):
  

    """
    The class contains parameters of Clusters and Neighbors CN noise model (currently it does not support neighbors). 
    
    It is used to:

    1) store results of CN noise model reconstruction algorithm (noise characterization routine)

    2) Mitigate redout errors on marginal probability distributions (error mitigation routine)  


    Parameters
    ----------
    clusters_tuple : a tuple of tuples 
        Stores information about clusters of qubits. Eeach tuple contains indices of qubits belonging to a cluster  e.g. ((0,2),(1,3))   

    noise_matrices: dictionary (key - tuple, value - array)
        Stores noise matrices associated to clusters. Key in the dictionary is a tuple with indices of qubits belonging to a cluster. 
        Corresponding value is an array storing noise matrix associated with the cluster. Eg. (0:2) : [[0.8,0,0,0], [0.2,1,0.1,0],[0,0,0.9,0],[0,0,0,1]].  

    inverse_noise_matrices : dictionary (key - tuple, value - array)
        Stores noise matrices associated to clusters. Key in the dictionary is a tuple with indices of qubits belonging to a cluster. 
        Corresponding value is an array storing inverse noise matrix associated with the cluster. Eg. (0:2) : [[0.8,0,0,0], [0.2,1,0.1,0],[0,0,0.9,0],[0,0,0,1]].
        This property is filled automatically when noise_matrices property is set via set_noise_matrices_dictionary method.

    qubit_in_cluster_membership: dictionary (key: tuple, value: tuple)
        Stores information about qubits membership in clusters. Key in the dictionary is a tuple constructed form a qubit index. Corresponding value is a cluster to which the qubit belongs.
        E.g. (2,) : (0,2,5). This property is filled automatically when noise_matrices property is set via set_noise_matrices_dictionary method

    
    Methods
    ----------
    set_noise_model:  
        Stores information about clusters of qubits. Eeach tuple contains indices of qubits belonging to a cluster  e.g. ((0,2),(1,3))   

    noise_matrices: dictionary (key - tuple, value - array)
        Stores noise matrices associated to clusters. Key in the dictionary is a tuple with indices of qubits belonging to a cluster. 
        Corresponding value is an array storing noise matrix associated with the cluster. Eg. (0:2) : [[0.8,0,0,0], [0.2,1,0.1,0],[0,0,0.9,0],[0,0,0,1]].  

    inverse_noise_matrices : dictionary (key - tuple, value - array)
        Stores noise matrices associated to clusters. Key in the dictionary is a tuple with indices of qubits belonging to a cluster. 
        Corresponding value is an array storing inverse noise matrix associated with the cluster. Eg. (0:2) : [[0.8,0,0,0], [0.2,1,0.1,0],[0,0,0.9,0],[0,0,0,1]].
        This property is filled automatically when noise_matrices property is set via set_noise_matrices_dictionary method.

    qubit_in_cluster_membership: dictionary (key: tuple, value: tuple)
        Stores information about qubits membership in clusters. Key in the dictionary is a tuple constructed form a qubit index. Corresponding value is a cluster to which the qubit belongs.
        E.g. (2,) : (0,2,5). This property is filled automatically when noise_matrices property is set via set_noise_matrices_dictionary method
    

'
"""
    def __init__(self, number_of_qubits: int):
        super().__init__()
        self.number_of_qubits = number_of_qubits
        self.locality = None  #integer
        self.clusters_tuple=()
        self.noise_matrices = {}
        self.inverse_noise_matrices ={}
        self.qubit_in_cluster_membership ={}
        self.composite_inverse_noise_matrices = {}
        self.clusters_neighborhoods = {}
    

    def _establish_qubit_in_cluster_dictionary(self):
        """
        Method used to establish membership of qubits in clusters. 
        It constructs a dictionary with key - tuple consisting of a qubit index e.g. (0,) and value - tuple storing corresponding cluster e.g. (0,3)
        It is automatically called when set_clusters_tuple method is used.     

        """
        for cluster in self.clusters_tuple:
            for qubit in cluster:
                self.qubit_in_cluster_membership[(qubit,)]= cluster
    
    def _compute_inverse_noise_matrices(self):
        """
        Method that creates inverse noise matrix dictionary for an instance of CN noise model. 
        In the dictionary key corresponds to a tuple storing a cluster, and value to an numpy array storing inverse noise matrix e.g. (0,5): 4 by 4 numpy array.
        It is automatically called when set_noise_matrices_dictionary is used.     

        """

        for cluster, noise_matrix in self.noise_matrices.items():
            
            if bool(self.clusters_neighborhoods):

                try:
                    self.inverse_noise_matrices[(cluster)] = np.linalg.inv(noise_matrix['averaged'])
                except:
                    print(f"MATRIX FOR SUBSET: {(cluster)} is not invertible, computing pseudinverse",'','red')
                    self.inverse_noise_matrices[(cluster)] = np.linalg.pinv(noise_matrix)['averaged']


            else:
                try:
                    self.inverse_noise_matrices[(cluster)] = np.linalg.inv(noise_matrix)
                except:
                    print(f"MATRIX FOR SUBSET: {(cluster)} is not invertible, computing pseudinverse",'','red')
                    self.inverse_noise_matrices[(cluster)] = np.linalg.pinv(noise_matrix)
    

    def set_clusters_tuple(self,clusters_tuple: tuple):
        """
        Method fills clusters_tuple property. It is called when properties of a CN noise model are established via set_noise_model method.   
        In addition the method calls _establish_qubit_in_cluster_dictionary, which fills qubit_in_cluster_membership property 

        """
        if isinstance(clusters_tuple,tuple) and isinstance(clusters_tuple[0],tuple):
            self.clusters_tuple = clusters_tuple
            self._establish_qubit_in_cluster_dictionary()
        else:
             raise Exception("Wrong data format. Cluster list should be a tuple of tuples")
    
    
        
    def set_noise_matrices_dictionary(self, noise_matrices_dictionary: Dict):

        """
        Method fills noise_matrices property. It is called when properties of a CN noise model are established via set_noise_model method.   
        In addition the method calls _compute_inverse_noise_matrices, which fills inverse_noise_matrices property.

        Parameters
        ----------
        noise_matrices_dictionary : dictionary (key - tuple of integers, value - array)
            Dictionary storing noise matrices for a CN noise model. Key in the dictionary is a tuple with indices of qubits belonging to a cluster. 
            Corresponding value is an array storing noise matrix associated with the cluster. Eg. (0:2) : [[0.8,0,0,0], [0.2,1,0.1,0],[0,0,0.9,0],[0,0,0,1]].
        
        

        """


        if isinstance(noise_matrices_dictionary,dict):
            self.noise_matrices =  noise_matrices_dictionary
            self._compute_inverse_noise_matrices()
        else:
             raise Exception("Wrong data format. Cluster list should be a list of tuples")
        
    def set_noise_model(self,noise_matrices_dictionary):

        """
        Main method used to fill properties of CN noise model. The following properties are filled when the method is called:    
        1) clusters_tuple
        2) qubit_in_cluster_membership
        3) noise_matrices 
        4) inverse_noise_matrices 
          

        Parameters
        ----------
        noise_matrices_dictionary : dictionary (key - tuple of integers, value - array)
            Dictionary storing noise matrices for a CN noise model. Key in the dictionary is a tuple with indices of qubits belonging to a cluster. 
            Corresponding value is an array storing noise matrix associated with the cluster. Eg. (0:2) : [[0.8,0,0,0], [0.2,1,0.1,0],[0,0,0.9,0],[0,0,0,1]].

        """

        self.set_clusters_tuple(clusters_tuple=tuple(noise_matrices_dictionary.keys()))
        self.set_noise_matrices_dictionary(noise_matrices_dictionary=noise_matrices_dictionary)

    def compute_extended_inverse_noise_matrices(self, clusters_in_marginal_list:List[Tuple])-> np.array:

        """
        Function determines inverse noise matrix for a given marginal. It is used as a step in mitigation_marginal function.
        The matrix is constructed as a tensor product on inverse noise matrices involved in a marginal, when it is necessary
        the matrix is permuted to ensure proper ordering of qubits indices. E.g. for a cluster_in_marginal_list 
        [(0,2),(1,3)], the returned matrix corresponds to qubits ordered as (0,1,2,3).   


        Parameters
        ----------
        noise_model : object of CNModelData class
            An object of CNModelData class 

        clusters_in_marginal_list:
            A list of tuples with clusters involved in the marginal   
            

        Returns
        -------

            An inverse noise matrix for qubits specified in clusters_in_marginal_list, qubits are sorted in ascending order 
            (e.g. for a clusters_in_marginal_list =[(0,4),(1,8)], indices of the inverse noise matrix indices correspond to qubits in the order (0,1,4,8) )
        
        """ 

        # a list of qubits is established
        unordered_qubits_in_marginal_list = list(reduce(operator.concat, clusters_in_marginal_list))

        #the inverse noise matrix is initialized as a unit numpy array 
        marginal_inverse_noise_matrix = np.array([1])
        
        # a loop over clusters in clusters_in_marginal_list
        for cluster in clusters_in_marginal_list:
        
            # total noise matrix is updated by taking tensor product with a inverse noise matrix of the currect cluster
            marginal_inverse_noise_matrix = np.kron(marginal_inverse_noise_matrix, self.inverse_noise_matrices[cluster])
    
        #final noise matrix is returned, permuted to preserve ascending order of qubit indices if necessary 
        marginal_inverse_noise_matrix = math.permute_composite_matrix(qubits_list=unordered_qubits_in_marginal_list,noise_matrix=marginal_inverse_noise_matrix)

        unordered_qubits_in_marginal_list.sort()

        key = tuple(unordered_qubits_in_marginal_list)

        self.inverse_noise_matrices[key] = marginal_inverse_noise_matrix 

        return marginal_inverse_noise_matrix 




#MOVE TO CNModelData class   

    def get_clusters_in_marginal_list(self, marginal : Tuple[int]) -> List[Tuple]:
        
        """
        Function creates a list of clusters that are involved in a marginal. Used in mitigation routines.
        For a given marginal inspects provided noise model and checks clusters membership of qubits form
        marginal.
        
        Parameters
        ----------
        marginal : tuple
            A tuple specifying marginal 

        noise_model : object of CNModelData class
            An object of CNModelData class  
            

        Returns
        -------

        clusters_in_marginal_list
            A list of tuples involved in the input marginal 
            
        
        """  

        #list storing clusters involved in the marginal, empty for start
        clusters_in_marginal_list=[]
        
        #a loop over qubits in the marginal
        for qubit in marginal:
            
            #a cluster to which the qubit belongs is determined 
            cluster = self.qubit_in_cluster_membership[(qubit,)]
            
            #this cluster is appended to clusters_in_marginal_list if it is not there 
            if cluster not in clusters_in_marginal_list:
                clusters_in_marginal_list.append(self.qubit_in_cluster_membership[(qubit,)])
        
        #results are returned
        return clusters_in_marginal_list
    

    def get_clusters_and_neighborhoods_in_marginal_dictionary(self, marginal : Tuple[int]) -> Dict[Tuple[int], Tuple[int]]:
        
        """
        Function creates a list of clusters that are involved in a marginal. Used in mitigation routines.
        For a given marginal inspects provided noise model and checks clusters membership of qubits form
        marginal.
        
        Parameters
        ----------
        marginal : tuple
            A tuple specifying marginal 

        noise_model : object of CNModelData class
            An object of CNModelData class  
            

        Returns
        -------

        clusters_in_marginal_list
            A list of tuples involved in the input marginal 
            
        
        """  

        #list storing clusters involved in the marginal, empty for start
        clusters_and_neighbors_in_marginal_dictionary={}
        
        if bool(self.clusters_neighborhoods):
        #a loop over qubits in the marginal
            for qubit in marginal:
                
                #a cluster to which the qubit belongs is determined 
                cluster = self.qubit_in_cluster_membership[(qubit,)]

                
                
                #this cluster is appended to clusters_in_marginal_list if it is not there 
                if cluster not in clusters_and_neighbors_in_marginal_dictionary.keys():
                    clusters_and_neighbors_in_marginal_dictionary[cluster] = self.clusters_neighborhoods[cluster]
            
            #results are returned
        return clusters_and_neighbors_in_marginal_dictionary
    


    def get_neighbors_in_marginal(self, marginal : Tuple[int]) -> Dict[Tuple[int], Tuple[int]]:
        
        """
        Function creates a dictionary with clusters and neighbors that belong to a marginal (it can happen that clusters' neighbors do not belong to other clusters that form a marginal .
        For a given marginal, and clusters and neighbors, membership of neighbors in the marginal is checked  
        
        Parameters
        ----------
        marginal : tuple
            A tuple specifying marginal 

        noise_model : object of CNModelData class
            An object of CNModelData class  
            

        Returns
        -------

        clusters_in_marginal_list
            A list of tuples involved in the input marginal 
            
        
        """  
        
        #here a dictionary is created to store neighbors that take part in a marginal 

        #dictionary with keys corresponding to clusters and values corresponding to neighbors that are in a marginal, and their position relative to a neighborhood 
        neighbors_coinciding_with_clusters_dictionary = {}

        clusters_and_neighbors_in_marginal_dictionary = self.get_clusters_and_neighborhoods_in_marginal_dictionary(marginal=marginal)

        qubits_in_marginal_list = list(reduce(operator.concat, list(clusters_and_neighbors_in_marginal_dictionary.keys())))


        #a loop over clusters and neighbors 
        for cluster, neighbors in clusters_and_neighbors_in_marginal_dictionary.items():

            # a temporary list storing neighbors
            neighbor_in_marginal = []

            #if there are neighbors for a cluster the check is performed 
            if neighbors!= None:

                #qubits in neighborhood are enumerated to store their relative position
                for index, qubit in enumerate(neighbors):

                    #if a qubit form this neighborhood is present in the marginal, its index and relative position is stored in a list 
                    if qubit in qubits_in_marginal_list:

                        neighbor_in_marginal.append((index,qubit))
                    
                #if after a previous loop neighbors were identified a dictionary entry is cerated
                if len(neighbor_in_marginal) > 0:

                    neighbors_coinciding_with_clusters_dictionary[cluster] = tuple(neighbor_in_marginal)
            
        return neighbors_coinciding_with_clusters_dictionary 
    

    def get_noise_matrices_indexes(self, marginal : Tuple[int] ,state):
        
        """
        Function creates a dictionary with indices of state dependent noise matrices that should be used to compute noise matrix inverse, 

        It is used during error  mitigation of CN noise model with neighbors.  
        
        Parameters
        ----------
        marginal : tuple
            A tuple specifying marginal 

        noise_model : object of CNModelData class
            An object of CNModelData class  
            

        Returns
        -------

        neighbors_indices_to_compute_average 
            A dictionary with a key - cluster and value - tuple of indicies to be averaged over or a string 'averaged' 
            
              
            
        
        """  
        
        
        
        #a dictionary to store information about noise matrices needed to compute an inverse for a given cluster is created 
        neighbors_indices_to_compute_average ={} 

        # this loop goes over all clusters that take part in a marginal, and their neighbors 

        neighbors_coinciding_with_clusters_dictionary = self.get_neighbors_in_marginal(marginal=marginal)

        clusters_and_neighbors_in_marginal_dictionary = self.get_clusters_and_neighborhoods_in_marginal_dictionary(marginal=marginal) 

        
        for cluster, neighbors in clusters_and_neighbors_in_marginal_dictionary.items():
            
            #this conditions verify whether given cluster has a neighbor that actual coincides with the marginal    
            if cluster in neighbors_coinciding_with_clusters_dictionary.keys():

                #here all possible states of the clusters neighbors are created 
                possible_states_list = list(itertools.product([0,1], repeat = len(neighbors)))
                
                #here we check, which states of neighbors are needed to compute inverse noise matrix 
                for neighbor_tuple in neighbors_coinciding_with_clusters_dictionary[cluster]:

                #possible_state is ordered according to position of a qubit in its neighborhood (relative index), while the position of a qubit in state is just its index
                    possible_states_list = [ possible_state for possible_state in possible_states_list if possible_state[neighbor_tuple[0]] == int(state[neighbor_tuple[1]]) ]
                
                neighbors_indices_to_compute_average[cluster] = tuple(possible_states_list)
            
            #if there are no neighbors in the marginal averaged noise matrix is used 
            else:

                neighbors_indices_to_compute_average[cluster] = 'averaged'
        
        return neighbors_indices_to_compute_average 


    def compute_extended_inverse_noise_matrices_state_dependent(self, clusters_in_marginal_list:List[Tuple],marginal,state)-> np.array:

        """
        Function determines inverse noise matrix for a given marginal. It is used as a step in mitigation_marginal function.
        The matrix is constructed as a tensor product on inverse noise matrices involved in a marginal, when it is necessary
        the matrix is permuted to ensure proper ordering of qubits indices. E.g. for a cluster_in_marginal_list 
        [(0,2),(1,3)], the returned matrix corresponds to qubits ordered as (0,1,2,3).   


        Parameters
        ----------
        noise_model : object of CNModelData class
            An object of CNModelData class 

        clusters_in_marginal_list:
            A list of tuples with clusters involved in the marginal   
            

        Returns
        -------

            An inverse noise matrix for qubits specified in clusters_in_marginal_list, qubits are sorted in ascending order 
            (e.g. for a clusters_in_marginal_list =[(0,4),(1,8)], indices of the inverse noise matrix indices correspond to qubits in the order (0,1,4,8) )
        
        """ 

        # a list of qubits is established
        unordered_qubits_in_marginal_list = list(reduce(operator.concat, clusters_in_marginal_list))

        #indicies of needed noise matrix are established
        noise_matrices_indicies = self.get_noise_matrices_indexes(marginal=marginal,state=state)

        #the inverse noise matrix is initialized as a unit numpy array 
        marginal_inverse_noise_matrix = np.array([1])
        
        # a loop over clusters in clusters_in_marginal_list
        for cluster in clusters_in_marginal_list:
            
            if noise_matrices_indicies[cluster]  == 'averaged':      
                
                # total noise matrix is updated by taking tensor product with a inverse noise matrix of the currect cluster
                marginal_inverse_noise_matrix = np.kron(marginal_inverse_noise_matrix, self.inverse_noise_matrices[cluster])

            else:

                current_noise_matrix = self.compute_specific_state_dependent_noise_matrix(cluster=cluster, indicies=noise_matrices_indicies[cluster])

                current_inverse_noise_matrix = np.linalg.inv(current_noise_matrix)

                marginal_inverse_noise_matrix = np.kron(marginal_inverse_noise_matrix, current_inverse_noise_matrix)




    
        #final noise matrix is returned, permuted to preserve ascending order of qubit indices if necessary 
        marginal_inverse_noise_matrix = math.permute_composite_matrix(qubits_list=unordered_qubits_in_marginal_list,noise_matrix=marginal_inverse_noise_matrix)

        unordered_qubits_in_marginal_list.sort()

        key = tuple(unordered_qubits_in_marginal_list)

        #self.inverse_noise_matrices[key] = marginal_inverse_noise_matrix 

        return marginal_inverse_noise_matrix 
    

    def compute_specific_state_dependent_noise_matrix(self, cluster: Tuple[int], indicies: Tuple[Tuple[int]]) -> np.array:

        noise_matrix = self.noise_matrices[cluster]
        
        normalization_factor = len(indicies)
        
        dim = len(noise_matrix['averaged'] )

        averaged_noise_matrix = np.zeros((dim,dim))

        for index in indicies:

            averaged_noise_matrix += noise_matrix[index]

        averaged_noise_matrix = 1/normalization_factor * averaged_noise_matrix

        return averaged_noise_matrix

        



    
   
# from qrem.functions_qrem import ancillary_functions as anf

# def t1(test_model:type[CNModelData]):
#     # Test of json export

    
#     json_dict_test = test_model.to_json()

#     new_model =CNModelData(number_of_qubits=5)

#     new_model.import_json(json_dict_test)

# def t2(test_model:type[CNModelData]):
#     # Test of pickle export

 
#     pickle_data = test_model.export_pickle("test_data")

#     new_model =CNModelData(number_of_qubits=5)

#     new_model.import_pickle("test_data")

#     print(new_model.clusters_tuple)

    


if __name__=='__main__':
   #set clusters tuple
    qubits_in_cluster_1=(0,3)
    qubits_in_cluster_2=(1,2,4)

    #generation noise matrice for clusters with built in function probability.random_stochastic_matrix
    noise_matrix_1= probability.random_stochastic_matrix(2**2)
    noise_matrix_2= probability.random_stochastic_matrix(2**3)
    
    noise_matrices_dictionary_test = {qubits_in_cluster_1: noise_matrix_1, qubits_in_cluster_2: noise_matrix_2}

    #create noise model 
    test_model= CNModelData(number_of_qubits=5)
    test_model.set_noise_model(noise_matrices_dictionary=noise_matrices_dictionary_test) 

    t2(test_model)
 

        


    #}



   # test_model.load_from_dict(dictionary_to_load)
   # print('after loading:\n', test_model.get_dict_format())

    #NOTE: Is it possible to have a nested dictionary with keys that are not string? 
    #json_dict_test = test_model.to_json()
 
    #pickle_dict_test = test_model.get_pickle()
    #test_model.export_pickle('exported_pickle', overwrite=True)
    #print('after change:\n', test_model.get_dict_format())
    #test_model.import_pickle('exported_pickle')


# -*- coding: utf-8 -*-
"""
The module contains helper functions for physics applications.

Created on Fri May 16 14:27:47 2014

@author: Peter Wittek
"""
from __future__ import division, print_function
from sympy.core import S
from sympy.physics.quantum.dagger import Dagger
from .nc_utils import generate_operators, flatten
from .solver_common import solve_sdp
from .sdp_relaxation import SdpRelaxation


def get_neighbors(index, lattice_length, width=0, periodic=False):
    """Get the forward neighbors of a site in a lattice.

    :param index: Linear index of operator.
    :type index: int.
    :param lattice_length: The size of the 2D lattice in either dimension
    :type lattice_length: int.
    :param width: Optional parameter to define width.
    :type width: int.
    :param periodic: Optional parameter to indicate periodic boundary
                     conditions.
    :type periodic: bool

    :returns: list of int -- the neighbors in linear index.
    """
    if width == 0:
        width = lattice_length
    neighbors = []
    coords = divmod(index, width)
    if coords[1] < width - 1:
        neighbors.append(index + 1)
    elif periodic and width > 1:
        neighbors.append(index - width + 1)
    if coords[0] < lattice_length - 1:
        neighbors.append(index + width)
    elif periodic:
        neighbors.append(index - (lattice_length - 1) * width)
    return neighbors


def get_next_neighbors(indices, lattice_length, width=0, distance=1,
                       periodic=False):
    """Get the forward neighbors at a given distance of a site or set of sites
    in a lattice.

    :param index: Linear index of operator.
    :type index: int.
    :param lattice_length: The size of the 2D lattice in either dimension
    :type lattice_length: int.
    :param width: Optional parameter to define width.
    :type width: int.
    :param distance: Optional parameter to define distance.
    :type width: int.
    :param periodic: Optional parameter to indicate periodic boundary
                     conditions.
    :type periodic: bool

    :returns: list of int -- the neighbors at given distance in linear index.
    """
    if not isinstance(indices, list):
        indices = [indices]
    if distance == 1:
        return flatten(get_neighbors(index, lattice_length, width, periodic)
                       for index in indices)
    else:
        s1 = set(flatten(get_next_neighbors(get_neighbors(index,
                                                          lattice_length,
                                                          width, periodic),
                                            lattice_length, width, distance-1,
                                            periodic) for index in indices))
        s2 = set(get_next_neighbors(indices, lattice_length, width, distance-1,
                                    periodic))
        return list(s1 - s2)


def bosonic_constraints(a):
    """Return  a set of constraints that define fermionic ladder operators.

    :param a: The non-Hermitian variables.
    :type a: list of :class:`sympy.physics.quantum.operator.Operator`.
    :returns: a dict of substitutions.
    """
    substitutions = {}
    for i, ai in enumerate(a):
        substitutions[ai * Dagger(ai)] = 1.0 + Dagger(ai) * ai
        for aj in a[i+1:]:
            # substitutions[ai*Dagger(aj)] = -Dagger(ai)*aj
            substitutions[ai*Dagger(aj)] = Dagger(aj)*ai
            substitutions[Dagger(ai)*aj] = aj*Dagger(ai)
            substitutions[ai*aj] = aj*ai
            substitutions[Dagger(ai) * Dagger(aj)] = Dagger(aj) * Dagger(ai)

    return substitutions


def fermionic_constraints(a):
    """Return  a set of constraints that define fermionic ladder operators.

    :param a: The non-Hermitian variables.
    :type a: list of :class:`sympy.physics.quantum.operator.Operator`.
    :returns: a dict of substitutions.
    """
    substitutions = {}
    for i, ai in enumerate(a):
        substitutions[ai ** 2] = 0
        substitutions[Dagger(ai) ** 2] = 0
        substitutions[ai * Dagger(ai)] = 1.0 - Dagger(ai) * ai
        for aj in a[i+1:]:
            # substitutions[ai*Dagger(aj)] = -Dagger(ai)*aj
            substitutions[ai*Dagger(aj)] = -Dagger(aj)*ai
            substitutions[Dagger(ai)*aj] = -aj*Dagger(ai)
            substitutions[ai*aj] = -aj*ai
            substitutions[Dagger(ai) * Dagger(aj)] = - Dagger(aj) * Dagger(ai)

    return substitutions


def pauli_constraints(X, Y, Z):
    """Return  a set of constraints that define Pauli spin operators.

    :param X: List of Pauli X operator on sites.
    :type X: list of :class:`sympy.physics.quantum.operator.HermitianOperator`.
    :param Y: List of Pauli Y operator on sites.
    :type Y: list of :class:`sympy.physics.quantum.operator.HermitianOperator`.
    :param Z: List of Pauli Z operator on sites.
    :type Z: list of :class:`sympy.physics.quantum.operator.HermitianOperator`.

    :returns: tuple of substitutions and equalities.
    """
    substitutions = {}
    n_vars = len(X)
    for i in range(n_vars):
        # They square to the identity
        substitutions[X[i] * X[i]] = 1
        substitutions[Y[i] * Y[i]] = 1
        substitutions[Z[i] * Z[i]] = 1

        # Anticommutation relations
        substitutions[Y[i] * X[i]] = - X[i] * Y[i]
        substitutions[Z[i] * X[i]] = - X[i] * Z[i]
        substitutions[Z[i] * Y[i]] = - Y[i] * Z[i]
        # Commutation relations.
        # equalities.append(X[i]*Y[i] - 1j*Z[i])
        # equalities.append(X[i]*Z[i] + 1j*Y[i])
        # equalities.append(Y[i]*Z[i] - 1j*X[i])
        # They commute between the sites
        for j in range(i + 1, n_vars):
            substitutions[X[j] * X[i]] = X[i] * X[j]
            substitutions[Y[j] * Y[i]] = Y[i] * Y[j]
            substitutions[Y[j] * X[i]] = X[i] * Y[j]
            substitutions[Y[i] * X[j]] = X[j] * Y[i]
            substitutions[Z[j] * Z[i]] = Z[i] * Z[j]
            substitutions[Z[j] * X[i]] = X[i] * Z[j]
            substitutions[Z[i] * X[j]] = X[j] * Z[i]
            substitutions[Z[j] * Y[i]] = Y[i] * Z[j]
            substitutions[Z[i] * Y[j]] = Y[j] * Z[i]
    return substitutions


def generate_measurements(party, label):
    """Generate variables that behave like measurements.

    :param party: The list of number of measurement outputs a party has.
    :type party: list of int.
    :param label: The label to be given to the symbolic variables.
    :type label: str.

    :returns: list of list of
             :class:`sympy.physics.quantum.operator.HermitianOperator`.
    """
    measurements = []
    for i in range(len(party)):
        measurements.append(generate_operators(label + '%s' % i, party[i] - 1,
                                               hermitian=True))
    return measurements


def projective_measurement_constraints(*parties):
    """Return a set of constraints that define projective measurements.

    :param parties: Measurements of different parties.
    :type A: list or tuple of list of list of
             :class:`sympy.physics.quantum.operator.HermitianOperator`.

    :returns: substitutions containing idempotency, orthogonality and
              commutation relations.
    """
    substitutions = {}
    # Idempotency and orthogonality of projectors
    if isinstance(parties[0][0][0], list):
        parties = parties[0]
    for party in parties:
        for measurement in party:
            for projector1 in measurement:
                for projector2 in measurement:
                    if projector1 == projector2:
                        substitutions[projector1**2] = projector1
                    else:
                        substitutions[projector1*projector2] = 0
                        substitutions[projector2*projector1] = 0
    # Projectors commute between parties in a partition
    for n1 in range(len(parties)):
        for n2 in range(n1+1, len(parties)):
            for measurement1 in parties[n1]:
                for measurement2 in parties[n2]:
                    for projector1 in measurement1:
                        for projector2 in measurement2:
                            substitutions[projector2*projector1] = \
                                projector1*projector2
    return substitutions


def define_objective_with_I(I, *args):
    """Define a polynomial using measurements and an I matrix describing a Bell
    inequality.

    :param I: The I matrix of a Bell inequality in the Collins-Gisin notation.
    :type I: list of list of int.
    :param args: Either the measurements of Alice and Bob or a `Probability`
                 class describing their measurement operators.
    :type A: tuple of list of list of
             :class:`sympy.physics.quantum.operator.HermitianOperator` or
             :class:`ncpol2sdpa.Probability`

    :returns: :class:`sympy.core.expr.Expr` -- the objective function to be
              solved as a minimization problem to find the maximum quantum
              violation. Note that the sign is flipped compared to the Bell
              inequality.
    """
    objective = I[0][0]
    if len(args) > 2 or len(args) == 0:
        raise Exception("Wrong number of arguments!")
    elif len(args) == 1:
        A = args[0].parties[0]
        B = args[0].parties[1]
    else:
        A = args[0]
        B = args[1]
    i, j = 0, 1  # Row and column index in I
    for m_Bj in B:  # Define first row
        for Bj in m_Bj:
            objective += I[i][j] * Bj
            j += 1
    i += 1
    for m_Ai in A:
        for Ai in m_Ai:
            objective += I[i][0] * Ai
            j = 1
            for m_Bj in B:
                for Bj in m_Bj:
                    objective += I[i][j] * Ai * Bj
                    j += 1
            i += 1
    return -objective


def correlator(A, B):
    """Correlators between the probabilities of two parties.

    :param A: Measurements of Alice.
    :type A: list of list of
             :class:`sympy.physics.quantum.operator.HermitianOperator`.
    :param B: Measurements of Bob.
    :type B: list of list of
             :class:`sympy.physics.quantum.operator.HermitianOperator`.

    :returns: list of correlators.
    """
    correlators = []
    for i in range(len(A)):
        correlator_row = []
        for j in range(len(B)):
            corr = 0
            for k in range(len(A[i])):
                for l in range(len(B[j])):
                    if k == l:
                        corr += A[i][k] * B[j][l]
                    else:
                        corr -= A[i][k] * B[j][l]
            correlator_row.append(corr)
        correlators.append(correlator_row)
    return correlators


def maximum_violation(A_configuration, B_configuration, I, level, extra=None):
    """Get the maximum violation of a two-party Bell inequality.

    :param A_configuration: Measurement settings of Alice.
    :type A_configuration: list of int.
    :param B_configuration: Measurement settings of Bob.
    :type B_configuration: list of int.
    :param I: The I matrix of a Bell inequality in the Collins-Gisin notation.
    :type I: list of list of int.
    :param level: Level of relaxation.
    :type level: int.

    :returns: tuple of primal and dual solutions of the SDP relaxation.
    """
    P = Probability(A_configuration, B_configuration)
    objective = define_objective_with_I(I, P)
    if extra is None:
        extramonomials = []
    else:
        extramonomials = P.get_extra_monomials(extra)
    sdpRelaxation = SdpRelaxation(P.get_all_operators(), verbose=0)
    sdpRelaxation.get_relaxation(level, objective=objective,
                                 substitutions=P.substitutions,
                                 extramonomials=extramonomials)
    solve_sdp(sdpRelaxation)
    return sdpRelaxation.primal, sdpRelaxation.dual


class Probability(object):

    def __init__(self, *args, **kwargs):
        """Class for working with quantum probabilities.

        :param *args: Input configurations for each parties
        :type *args: tuple of list of lists
        :param labels: Optional parameter string to define the label of each
                       party.
        :type labels: list of str.
        :Example:

        For a CHSH scenario, instantiate the class as

            P = Probability([2, 2], [2, 2])

        """

        self.n_parties = len(args)
        self.parties = []
        self.labels = [chr(ord('A') + i) for i in range(self.n_parties)]
        for name, value in kwargs.items():
            if name == "labels":
                if len(value) != self.n_parties:
                    raise Exception("Incorrect number of labels!")
                else:
                    self.labels = value
            else:
                raise Exception("Unknown parameter " + name)
        for i, configuration in enumerate(args):
            self.parties.append(generate_measurements(configuration,
                                                      self.labels[i]))
        self.substitutions = projective_measurement_constraints(self.parties)

    def get_all_operators(self):
        """Return all operators across all parties and measurements to supply
        them to the `ncpol2sdpa.SdpRelaxation` class.

        """
        return flatten(self.parties)

    def _monomial_generator(self, monomials, label_indices):
        if label_indices == []:
            return monomials
        elif monomials == []:
            return self._monomial_generator(
                flatten(self.parties[label_indices[0]]), label_indices[1:])
        else:
            result = [m1*m2 for m1 in monomials
                      for m2 in flatten(self.parties[label_indices[0]])]
            return self._monomial_generator(result, label_indices[1:])

    def get_extra_monomials(self, *args):
        if len(args) == 0:
            return []
        if isinstance(args[0], list):
            args = args[0]
        extra_monomials = []
        for s in args:
            label_indices = [self.labels.index(party) for party in s]
            extra_monomials.extend(self._monomial_generator([], label_indices))
        return extra_monomials

    def _convert_marginal_index(self, marginal):
        if isinstance(marginal, str):
            return [self.labels.index(marginal)]
        else:
            return sorted([self.labels.index(m) if isinstance(m, str) else m
                           for m in marginal])

    def __call__(self, output_, input_, marginal=None):
        """Obtain your probabilities in the p(ab...|xy...) notation.

        :param output_: Conditional output as [a, b, ...]
        :type output_: list of ints.
        :param input_: The input to condition on as [x, y, ...]
        :type input_: list of ints.
        :param marginal: Optional parameter. If it is a marginal, then you can
                         define which party or parties it belongs to.
        :type marginal: list of str.
        :returns: polynomial of `sympy.physics.quantum.HermitianOperator`.

        :Example:

        For the CHSH scenario, to get p(10|01), write

            P([1,0], [0,1])

        To get the marginal p_A(0|1), write

            P([0], [1], ['A'])

        """

        if len(output_) != len(input_):
            raise Exception("The number of inputs does not match the number of"
                            "outputs!")
        elif len(input_) > self.n_parties:
            raise Exception("The number of inputs exceeds the number of "
                            "parties!")
        elif marginal is None and len(input_) < self.n_parties:
            raise Exception("Marginal requested, but without defining which!")
        elif marginal is None:
            marginal = self._convert_marginal_index(self.labels)
        else:
            marginal = self._convert_marginal_index(marginal)
            if len(marginal) != len(input_):
                raise Exception("The number of parties in the marginal does "
                                "not match the number of inputs!")
        result = S.One
        for party, (proj, meas) in enumerate(zip(output_, input_)):
            if len(self.parties[marginal[party]]) < meas + 1:
                raise Exception("Invalid measurement index " + str(meas) +
                                " for party " + self.labels[party])
            elif len(self.parties[marginal[party]][meas]) < proj:
                raise Exception("Invalid projection operator index " +
                                str(proj) + " for party " + self.labels[party])
            elif len(self.parties[marginal[party]][meas]) == proj:
                # We are in the Collins-Gisin picture: the last projector
                # is not part of the measurement.
                result *= S.One - \
                    sum(op for op in self.parties[marginal[party]][meas])
            else:
                result *= self.parties[marginal[party]][meas][proj]
        return result



"""qrem.common.math module contains helper functions, that can be used to save and load 
intermediate or final results.
"""

from datetime import datetime
from typing import List, Dict, Union, Optional, Callable, Tuple

from pathlib import Path
import shutil
from qrem.common import printer
import pickle
#===========================
# Helpers for dates in path safe mode
#===========================

def date_time_formatted():
    return datetime.now().strftime("%Y_%m_%d-%H_%M_%S")
    
def date_formatted():
    return datetime.now().strftime("%Y-%m-%d")

def date_time_fileformatted():
    return datetime.now().strftime("%Y_%m_%d__%H_%M_%S")


def prepare_outfile(outpath: Union[bool,str,Path]=False,
                    overwrite: bool = False,
                    default_filename = "output.pkl"):
    

    if isinstance(outpath, str):
        backup = Path(outpath) 
    if isinstance(outpath, bool):
        backup = Path.home().joinpath("qrem_results") #makes a dir path
        
    if isinstance(outpath, Path):
        export_path = outpath
        if not export_path.exists() or export_path.is_dir():
            export_path.mkdir(parents=True, exist_ok=True)
            export_path.joinpath(default_filename)
        if export_path.is_file() :
            export_path.parent.mkdir(parents=True, exist_ok=True)  
    else: 
        export_path =Path.home().joinpath("qrem_results").joinpath(default_filename)
        printer.errprint("Wrong outpath type for backing up data")

    if not overwrite:
        assert (not export_path.exists())

    return export_path
#(PP) delete not used - maybe better than above?
def load(file_path: str):
    """
    Quickly load data from a pickled file containing a dictionary.
    """
    import pickle

    if file_path[-4:] != '.pkl':
        file_path = file_path + '.pkl'
    try:
        with open(file_path, 'rb') as f:
            unpickled = pickle.load(f)
    except(ValueError):
        with open(file_path, 'rb') as f:
            unpickled = pickle.load(f)

    return unpickled





#MOVE_TO >> core.io TO DELETE PP CAN BE A MARKER FOR UNUSED STUFF
def save(dictionary_to_save: dict,
            directory: Union[Path,str],
            custom_filename: Optional[str] = None,
            overwrite: bool = False,
            verbose_log: bool = True):
    """
    Quickly save data to a pickled file containing a dictionary.
    """
    
    if isinstance(directory,str):
        directory = Path(directory)

   


    if directory :
        dir = Path(directory) 
        dir.mkdir(mode=0o777, exist_ok=True, parents= True) 
    else:
        dir = Path.home().joinpath("qrem_results")
        dir.mkdir(mode=0o777, exist_ok=True) 

    if verbose_log:
        printer.qprint("Saving pickled file to: ", str(dir))
    
    # original_umask = os.umask(0)
    # os.umask(original_umask)


    if dir.is_dir():
        if any(dir.iterdir()) and overwrite:
            printer.warprint("Warning: overwriting folder", f"content of the folder {dir} has been overwritten")
            shutil.rmtree(dir)
            dir.mkdir(mode=0o777, exist_ok=False, parents= True) 
        elif not any(dir.iterdir()) and overwrite:
            printer.errprint("Error: folder exists", f"content of the folder {dir} has been overwritten")
            return
        

    if custom_filename is None:
        dt = date_time_fileformatted()
        file_path = dir.joinpath('Results_Object' + '___' + str(dt) + '.pkl')
    else:
        if custom_filename.split(".")[-1] != '.pkl':
            custom_filename = custom_filename + '.pkl'
        file_path = dir.joinpath(custom_filename)


    with open(file_path, 'wb') as f:
        pickle.dump(dictionary_to_save, f, pickle.HIGHEST_PROTOCOL)


"""
@authors: Filip Maciejewski, Oskar Słowik, Tomek Rybotycki
@contact: filip.b.maciejewski@gmail.com
"""

import copy

from qrem.functions_qrem import povmtools
from qrem.common import convert

def get_neighborhood_treshold_statitsical_pairs(number_of_samples,
                                                number_of_qubits=1,
                                                probability_of_error=0.001):
    pairs_number = number_of_qubits * (number_of_qubits - 1) / 2
    eps1q = povmtools.get_statistical_error_bound(2,
                                                  number_of_samples,
                                                  probability_of_error,
                                                  pairs_number)

    return 2 * eps1q


# def rename_clusters(clusters, true_indices):
#     proper_clusters = {}
#     for key, value in clusters.items():
#         indices_enum_key = convert.get_qubit_indices_from_keystring(key)
#         indices_enum_value = value

#         key_true = ''.join(['q%s' % true_indices[qi] for qi in indices_enum_key])
#         value_true = [true_indices[qi] for qi in indices_enum_value]

#         proper_clusters[key_true] = value_true

#     return proper_clusters


def cut_subset_sizes(clusters_neighbourhoods_dict,
                     correlations_table,
                     target_size=5):
    cutted_dict = copy.deepcopy(clusters_neighbourhoods_dict)

    for cluster, neighbours in clusters_neighbourhoods_dict.items():
        correlations_now = []
        cluster_inds = convert.get_qubit_indices_from_keystring(cluster)
        for ni in neighbours:
            for ci in cluster_inds:
                correlations_now.append([ni, correlations_table[ci, ni]])

        sorted_correlations = sorted(correlations_now, key=lambda x: x[1], reverse=True)

        base_size = len(cluster_inds)

        cut_neighbourhood = []

        if base_size == target_size:
            pass
        else:
            for tup in sorted_correlations:
                if (base_size + len(cut_neighbourhood)) == target_size:
                    # print()
                    break
                else:
                    ni = tup[0]
                    if ni not in cut_neighbourhood:
                        cut_neighbourhood.append(ni)

        cutted_dict[cluster] = sorted(cut_neighbourhood)

    return cutted_dict



"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""



from pathlib import Path
import configargparse
import ast
import os

from qrem.common.printer import qprint, errprint, warprint
#----------------
# Helper functions for parsing ini config files 
#----------------
def parse_str(in_str):
    'Helper function to evaluate string in config file'
    # could be a straight string, e.g 'string', or (double) quoted string "'string'", depending on the version of configargparse.
    if in_str[0] == '"': # double quoted
        return in_str.replace('"', '')
    elif in_str[0] == "'": # double quoted
        return in_str.replace("'", '')
    else:
        return in_str   
def parse_literal(instr):
    """ converts a string specifying a list """
    if instr is None or len(instr) == 0:
        return None
    print(instr)
    ret = ast.literal_eval(instr)
    print(ret)
    if isinstance(ret, str): # it was a list within a list
        print(ret)
        ret = ast.literal_eval(ret)
    return ret

def parse_boolean(b):
    """ Interpret various user inputs as boolean values """
    if len(b) < 1:
        raise ValueError('Cannot parse empty string into boolean.')
    b = b[0].lower()
    if b in ('t', 'y', '1'):
        return True
    if b in ('f', 'n', '0'):
        return False
    raise ValueError(f'Cannot interprete {b} as a boolean.')


#----------------
# path to local folder that contains find default config
# Order of precedence, from most important:
# - command line args / args added from config
# - environment variables 
# - config file values 
# - defaults defined in code
#----------------
common_path = Path(__file__).resolve().parent 

class QremConfigLoader:
    config_parser = None

    def load(cmd_args=None, default_path=str(Path(common_path,'default.ini')) , as_dict=False):
        """ processes config file specific for QREM """

        common_path = Path(__file__).resolve().parent   

        # If we receive a relative path, it is relative to this file location, not the cwd which could vary, e.g. if you run it with hython
        if os.path.isabs(default_path):
            cfg_path = default_path
        else:
            FILEPATH = os.path.dirname(os.path.realpath(__file__))
            cfg_path = os.path.join(FILEPATH, default_path)

        assert os.path.exists(cfg_path) #this should go in the final version of code

        QremConfigLoader.config_parser = configargparse.ArgParser(default_config_files=[cfg_path],
                                                description='QREM run config')
        
        # First, we get existing config file and add 
        QremConfigLoader.config_parser.add('-c', '--config_file', required=False, is_config_file=True, 
                        help='external config file path')
        
        #----
        #[general]
        # experiment_name and author will be saved in all of the files to be able to track experiment authors. can be a nick or any other tag
        QremConfigLoader.config_parser.add_argument('--experiment_name', type=parse_str, required=True, default='Example Experiment',
                                help='Experiment name will be saved in all of the filesfor bookeeping.')
        QremConfigLoader.config_parser.add_argument('--author', type=parse_str, required=True, default='Anonymous',
                                help='Author name/nick that will be saved in all of the filesfor bookeeping.')
        QremConfigLoader.config_parser.add_argument('--verbose_log', type=parse_boolean, required=False, default=False,
                                help='Turn on verbose logging for more printouts with info')

        #----
        #[data]
        QremConfigLoader.config_parser.add_argument('--backup_circuits', type=parse_boolean, required=False, default=True,
                                help='Whether to save original list of circuits in QREM definition to a file in experiment folder')
        QremConfigLoader.config_parser.add_argument('--backup_job_ids', type=parse_boolean, required=False, default=True,
                                help='Whether to save job ids to a file in experiment folder  after sending data to a backend ')
        QremConfigLoader.config_parser.add_argument('--backup_circuits_metadata', type=parse_boolean, required=False, default=False,
                                help='Whether to add more information about qubits and circuits to saved circuits list (like qubit readout errors list, list of bad qubits etc)')

        #----
        #[experiment]
        QremConfigLoader.config_parser.add_argument('--experiment_path', type=parse_str, required=True, default='C:\\experiments\\experiment_example\\',
                                help='Path to save all the expreiment inputs and results')
        QremConfigLoader.config_parser.add_argument('--device_name', type=parse_str, required=True, default='ASPEN-M-2',
                                help='Name of the device to run experiment on. Available names: ASPEN-M-2, TODO FILL IN')
        QremConfigLoader.config_parser.add_argument('--provider', type=parse_str, required=True, default='AWS-BRAKET',
                                help='Chosen provider. Available: AWS-BRAKET, IBM, TODO: FILL IN')        
        QremConfigLoader.config_parser.add_argument('--ibm_connection_method', type=parse_str, required=True, default='RUNTIME',
                                help='Valid only for IBM - how to connect to the quantum machine, available: RUNTIME_SESSIONS, RUNTIME, PROVIDER, JOB_EXECUTE, DEBUG')
        QremConfigLoader.config_parser.add_argument('--provider_instance', type=parse_str, required=True, default='AWS-BRAKET',
                                help='Valid only for IBM -  Provider instance, used values can be ibm-q/open/main or ... or ibm-q-psnc/internal/reservations')        

        QremConfigLoader.config_parser.add_argument('--experiment_type', type=parse_str, required=True, default='QDoT',
                                help='Chosen experiment type. Available values: QDoT, DDoT, QDT, RFE')
        QremConfigLoader.config_parser.add_argument('--k_locality', type=int, required=False, default=2,
                                help='k_locality value. Currently supported between 2-5')
        QremConfigLoader.config_parser.add_argument('--gate_threshold', type=float, required=False, default=0.005,
                                help='Assumed gate threshold for cut-off for qubits we want to take into account.')
        QremConfigLoader.config_parser.add_argument('--qbits_of_interest', required=False, type=parse_literal, default=None,
                                help='List of specific qubits to target for the experiment.')      
        QremConfigLoader.config_parser.add_argument('--limited_circuit_randomness', required=False, type=parse_boolean, default=True,
                                help='should any limitation of circuit randomness be imposed (e.g. limit on nubmer of random circuits/number of shots)')
        QremConfigLoader.config_parser.add_argument('--random_circuits_count', required=False, type=int, default=None,
                                help='count of random circuits to be sent (total = number of random circuits * how many times we repeat one)')
        QremConfigLoader.config_parser.add_argument('--shots_per_circuit', required=True, type=int, default=None,
                                help='How many times each random circuit should be repeated')
        QremConfigLoader.config_parser.add_argument('--aws_braket_task_retries', required=False, type=int, default=3,
                                help='How many times each aws task should be repeated')
        QremConfigLoader.config_parser.add_argument('--job_tags', required=False, type=parse_literal, default=["QREM_JOB",],
                                help='List of ags for the jobs send to the backend.')
        #QremConfigLoader.config_parser.add_argument('--run_on_all_qbits', required=False, type=parse_boolean, default=True)

        QremConfigLoader.config_parser.add_argument('--ensure_completnes_on_pairs', required=False, type=parse_boolean, default=True,
                                help='should we perform algorighmic drawing of circuits with complete pairs')
        
        QremConfigLoader.config_parser.add_argument('--ground_states_circuits', required=False, type=parse_boolean, default=True,
                                help='should a pre-prepared list of circuits of ground states of hamiltonians be added to random circuits')
        QremConfigLoader.config_parser.add_argument('--number_of_ground_states', type=int, required=False, default=10,
                                help='how many ground state circuits to add')
        QremConfigLoader.config_parser.add_argument('--ground_states_circuits_path', type=parse_str, required=False, default=None,
                                help='Path to pickle file with dictionary of int:circuit of qrem_circuits containing ground states')
        QremConfigLoader.config_parser.add_argument('--aws_pickle_results', type=parse_boolean, required=False, default=True,
                                help='Should we pickle results for AWS submission or calculate circuits on machine?')
        
        QremConfigLoader.config_parser.add_argument('--coherence_witness_circuits', type=parse_boolean, required=False, default=None,
                                help='Do circuits testing POVMs coherence should be included in the experiment? ')
        
        QremConfigLoader.config_parser.add_argument('--coherence_witness_circuits_path', type=parse_str, required=False, default=None,
                                help='Path to coherence witness circuits collection')
        


        config, unknown  = QremConfigLoader.config_parser.parse_known_args(cmd_args)


        #print(options)

        #If you want config as dictionary, make all the keys lower letters
        if as_dict:
            # cfg_dict_orig = config.__dict__.copy()
            # cfg_dict = config.__dict__
            # for key, val in cfg_dict_orig.items():
            #     cfg_dict[key] = val.lower()
            return config.__dict__
    
        return config
    
    def values():
        """ Return summary of configuration values """
        print("----------")
        if QremConfigLoader.config_parser is not None:
            qprint("Configuration values:")
            print(QremConfigLoader.config_parser.format_values()) 
        else:
            errprint("ERROR: Config not defined") 

    def help():
        """ Print help for preparation of the config file """
        print("----------")
        if QremConfigLoader.config_parser is not None:
            qprint("Configuration help:")
            print(QremConfigLoader.config_parser.format_help()) 
        else:
            errprint("ERROR: Config not defined") 




if __name__ == "__main__":
    aa = QremConfigLoader.load()
    QremConfigLoader.help()
    QremConfigLoader.values()

    # for key, val in cfg_dict_orig.items():
    #     cfg_dict[key] = val.lower()
    print(aa)


"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""
#TODO FBM: think whether do something here. For now, check noise_characterization.data_analysis.InitialNoiseAnalyzer
#ORGANIZE - this file seams obsolete (MO)


from typing import List, Dict, Union, Optional

import numpy as np
from tqdm import tqdm

from qrem.functions_qrem import povmtools
from qrem.functions_qrem.povmtools import reorder_probabilities
from qrem.types.circuit_collection import CircuitCollection


# ORGANIZE: this needs exception handling and securing against bad data
# ORGANIZE: write setters for all attributes or converter modules will handle this?
# ORGANIZE: the functions for converting between formats are copied from qiskit_utilities
# TEST: all of this.

class ExperimentResults(CircuitCollection):
    """
    The class contains data about results of an experiment. Qubit and classical outcome ordering - big endian: 0123.
    Extends qrem.types.CircuitCollection, so it contains description of circuits ran in the experiment (type of
            experiment, prepared states, used qubits, date etc.) - see class' description for details

 
    Parameters
    ----------
    frequencies (np.ndarray(c, 2^q)): array, where c is # of circuits, q is # of qubits, each entry contains
            frequency of a given state occurring in experiment result for given circuit.
            Possible states in a row are numbered increasingly from |00000 ... 0>, |10000 ... 0> up to |1111 ... 1>,
            Example: [[0.25, 0.00, 0.30, 0.45] ,
                      [0.00, 0.50, 0.50, 0.00]] is a result of 2 circuits on 2 qubits,
                      where in the first circuit outcome |00> occurred in 25% of shots, |10> in 0%, |01> in 0.30 etc.#TODO check if this is how we've used it so far
    
    original_counts: dict
        The key is the name of a circuit (str), the value is a dictionary of counts, where the key is a
        string denoting classical outcome and the value the number of its occurrences in given experiment.
        The circuits are named according to convention:
        circuit_name = "experiment name" + "-" + "circuit label" +
        "no" + "integer identifier for multiple implementations of the same circuit", e.g.: "DDOT-010no3"

    counts: dict
        The key is the label of a circuit (str), the value is a tuple containing all information about counts, according to following schematic:

        counts = {"<circuit_label>": tuple(<RESTULS_MATRIX>, <COUNTS_VECTOR>)}

        <circuit_label>: str, BigEndian-0 on left
        <RESTULS_MATRIX>: np.array(m x n),dtype=bool, "BigEndian-like" -first on left )
        <COUNTS_VECTOR>: np.array(m x 1),dtype=int)
        


    probabilities_list (List): TODO: find in code and comment what this is
    shots_results (List[List[str]] or np.ndarray(c, s, q)): measured outcome for each shot in each circuit, where
            c is # of circuits, s is # of shots in each circuit, q is # of qubits. Outcome saved as bitstring.
            Example: [['01001', '00100', '10101'],
                      ['01011', '11101', '01100']] is a result of 2 circuits on 5 qubits, 3 shots each circuit.
    tasks_IDs (str): device-specific ID for each circuit in the experiment, ordered #TODO - order how?
    datetime_obtained_utc: date of obtaining experiment results

    """

    def __init__(self):
        super().__init__()

        self.counts = None
        self.frequencies = None

        self._frequencies = None
        self.probabilities_list = None
        self.shots_results = None

        self.tasks_IDs = None
        #self.datetime_obtained_utc = None MOVE/SAVE TO self.metadata["resuts_obtained_date"] in utc format
    @property
    def frequencies(self): 
        print("getter method called") 
        if not self._frequencies:
            if self.counts:
                #fill from counts
                pass
            pass
        return self._frequencies 
    

    def load_from_dict(self, dictionary):
        for key in dictionary:
            if key in self.get_dict_format():
                setattr(self, key,
                        dictionary[key])  # TODO: secure against bad data
        pass

    def _get_counts_from_frequencies_array(self):
        if self.circuit_collection is not None:
            if self.frequencies_array is None:
                raise Exception("Frequencies_array is empty")
            else:
                pass
        else:
            raise Exception("Circuit_collection is empty - not enough data to generate counts")

    def _get_counts_from_shots_results(self):
        if self.circuit_collection is not None:
            if self.shots_results is None:
                raise Exception("Shots_results are empty")
            else:
                pass
        else:
            raise Exception("Circuit_collection is empty - not enough data to generate counts")

    def _get_counts_from_probabilities_list(self):
        if self.circuit_collection is not None:
            if self.probabilities_list is None:
                raise Exception("Probabilities_list is empty")
            else:
                pass
        else:
            raise Exception("Circuit_collection is empty - not enough data to generate counts")

    def _get_frequencies_array_from_counts(self):
        if self.counts is None:
            raise Exception("Counts are empty")
        else:
            pass

    def _get_frequencies_array_from_shots_results(self):
        if self.shots_results is None:
            raise Exception("Shots_results are empty")
        else:
            pass

    def _get_frequencies_array_from_probabilities_list(self):
        if self.probabilities_list is None:
            raise Exception("Probabilities_list is empty")
        else:
            pass

    def _get_probabilities_list_from_counts(self):
        if self.counts is None:
            raise Exception("Counts are empty")
        else:
            pass

    def _get_probabilities_list_from_shots_results(self):
        if self.shots_results is None:
            raise Exception("Shots_results are empty")
        else:
            pass

    def _get_probabilities_list_from_frequencies_array(self):
        if self.probabilities_list is None:
            raise Exception("Frequencies_array is empty")
        else:
            pass

    def get_counts(self):
        if self.counts is not None:
            return self.counts
        elif self.circuit_collection is not None:
            if self.frequencies_array is not None:
                return self._get_counts_from_frequencies_array()
            elif self.probabilities_list is not None:
                return self._get_counts_from_probabilities_list()
            elif self.shots_results is not None:
                return self._get_counts_from_shots_results()
        else:
            raise Exception("Not enough data to generate counts")

    def get_frequencies_array(self):
        if self.frequencies_array is not None:
            return self.frequencies_array
        elif self.counts is not None:
            return self._get_frequencies_array_from_counts()
        elif self.probabilities_list is not None:
            return self._get_frequencies_array_from_probabilities_list()
        elif self.shots_results is not None:
            return self._get_frequencies_array_from_shots_results()
        else:
            raise Exception("Not enough data to generate frequencies_array")

    def get_probabilities_list(self):
        if self.probabilities_list is not None:
            return self.probabilities_list
        elif self.counts is not None:
            return self._get_probabilities_list_from_counts()
        elif self.frequencies_array is not None:
            return self._get_probabilities_list_from_frequencies_array()
        elif self.shots_results is not None:
            return self._get_probabilities_list_from_shots_results()
        else:
            raise Exception("Not enough data to generate probabilities_list")

    def get_shots_results(self):
        if self.shots_results is not None:
            return self.shots_results
        else:
            raise Exception("Shots_results is empty")




    # ORGANIZE: The three methods below are copied functions from qiskit_utilities. Refactor
    def get_frequencies_array_from_probabilities_list(self, frequencies_list: List[Union[List[float], np.ndarray]],
                                                      reverse_order: Optional[bool] = False):
        """
        :param frequencies_list: list of probabilities estimated in experiments
        :param reverse_order: specify whether probabiltiies lists should be reordered, which corresponds
                              to changing qubits' ordering
        """

        number_of_qubits = int(np.log2(len(frequencies_list[0])))
        frequencies_formatted = frequencies_list
        if reverse_order:
            frequencies_formatted = [reorder_probabilities(probs, range(number_of_qubits)[::-1])
                                     for probs in frequencies_list]

        frequencies_array = np.ndarray(shape=(len(frequencies_list), len(frequencies_list[0])))

        for probe_state_index in range(len(frequencies_list)):
            frequencies_array[probe_state_index][:] = frequencies_formatted[:]
        return frequencies_array


    def get_frequencies_from_counts(counts_dict,
                                    crs=None,
                                    classical_register=None,
                                    shots_number=None,
                                    reorder_bits=True):
        if crs is None:
            crs = len(list(list(counts_dict.keys())[0]))

        d = 2 ** crs

        if classical_register is None:
            classical_register = ["{0:b}".format(i).zfill(crs) for i in range(d)]

        normal_order = []

        for j in range(d):
            if classical_register[j] in counts_dict.keys():
                counts_now = counts_dict[classical_register[j]]
                normal_order.append(counts_now)

            else:
                normal_order.append(0)
        if reorder_bits:
            frequencies = reorder_probabilities(normal_order, range(crs)[::-1])
        else:
            frequencies = normal_order

        if shots_number is None:
            frequencies = frequencies / np.sum(frequencies)
        else:
            frequencies = frequencies / shots_number

        return frequencies

    def add_counts_dicts(all_counts, modulo, dimension):
        frequencies = [np.zeros(dimension) for i in range(modulo)]

        for counts_index in tqdm(range(len(all_counts))):
            true_index = counts_index % modulo

            freqs_now = povmtools.counts_dict_to_frequencies_vector(all_counts[counts_index], True)
            frequencies[true_index][:] += freqs_now[:]

            # print(freqs_now)
        for i in range(modulo):
            frequencies[i] *= 1 / np.sum(frequencies[i])

        return frequencies



from qrem.ctmp.simtools.sampling_utils import check_transition, apply_transition
from qrem.types import CTMPModelData
from typing import Dict, Tuple, List
import numpy as np
import random


def generate_sample(CTMP_model: CTMPModelData,
                    input_state: str,
                    t: float=1) -> str:
    """Generates a sample according to the stochastic matrix exp(Gt) of the  CTMP model 
    acting on a classical input state by using Gillespie's algorithm. Default t=1.
    """
    if len(input_state) != CTMP_model.n:
        raise Exception("Input state size and model size do not match")
    time = 0
    next_state = input_state
    while time < t:
        current_state = next_state
        # check which transitions are available in the current state
        transitions = list(filter(lambda x: check_transition(current_state, x), CTMP_model.rates))
        if len(transitions) == 0:
            # stuck in a state where no transitions are available
            time = t
        else:
            # Gillespie step
            transition_rates = [error[4] for error in transitions]
            total_rate = sum(transition_rates)
            jump = np.random.exponential(1 / total_rate)
            time += jump
            # pick random transition
            k = random.choices(range(len(transition_rates)), [x/total_rate for x in transition_rates])[0]
            next_state = apply_transition(current_state, transitions[k])
    return current_state



from collections import Counter
from collections import defaultdict
from typing import Dict, Tuple

import numpy as np
from tqdm import tqdm

from qrem.functions_qrem import ancillary_functions as anf, functions_data_analysis as fda

from qrem.common.printer import qprint
from qrem.common import convert, utils


#MOcomm - porely documented file -learn what is inside


def __format_counts_from_clusters(counts_dict,
                                  clusters_list):
    number_of_qubits = np.max([np.max(cl) for cl in clusters_list]) + 1
    # print(number_of_qubits)

    number_of_clusters = len(clusters_list)
    counts_dict_formatted = {}
    for sample_list, number_of_times in counts_dict.items():
        output_bitstring_now = ["-" for _ in range(number_of_qubits)]
        for cluster_index in range(number_of_clusters):
            cluster = clusters_list[cluster_index]
            cluster_sample = sample_list[cluster_index]
            local_bitstring_now = convert.integer_to_bitstring(
                integer=int(np.argmax(cluster_sample)),
                number_of_bits=len(cluster))
            # print(local_bitstring_now,cluster,len(cluster))
            for qubit_index in range(len(cluster)):
                output_bitstring_now[cluster[qubit_index]] = local_bitstring_now[qubit_index]

        counts_dict_formatted["".join(output_bitstring_now)] = number_of_times
    return counts_dict_formatted


def sample_from_column_of_stochastic_map(column_bitstring,
                                         stochastic_map,
                                         number_of_samples=1):
    number_of_qubits = len(list(column_bitstring))
    distro = stochastic_map[:, int(column_bitstring, 2)]
    samples = np.random.multinomial(pvals=distro,
                                    n=number_of_samples)

    counts_dictionary = {}
    for integer in range(len(samples)):
        counts_dictionary[convert.integer_to_bitstring(integer=integer,
                                                   number_of_bits=number_of_qubits)] = integer

    return samples


def sample_from_multiple_input_states(input_states_dictionary: Dict[str, int],
                                      local_stochastic_matrices,
                                      print_progres_bar=False):
    counts_dicts_list = []

    items_range = input_states_dictionary.items()

    if print_progres_bar:
        items_range = tqdm(items_range)
    #
    # local_columns_big_matrices = {input_bitstring: {cluster:stochastic_map[:, int("".join([input_bitstring[x] for x in cluster]), 2)] for cluster, stochastic_map in
    #                               local_noise_matrices.items()} for input_bitstring in input_states_dictionary.keys()}

    for input_state, number_of_samples in items_range:
        counts_dicts_list.append(sample_from_product_noise_model(input_bitstring=input_state,
                                                                 number_of_samples=number_of_samples,
                                                                 local_noise_matrices=local_stochastic_matrices,
                                                                 # local_columns_big_dictionary=local_columns_big_matrices
                                                                 ))

    total_counts_dictionary = fda.merge_multiple_counts_dictionaries(counts_dicts_list)
    return total_counts_dictionary


def __format_counts_temporary(all_samples,
                              clusters_tuple,
                              number_of_qubits,
                              counts_dictionary_now):


    samples_pairs = list(zip(*all_samples))
    counting = Counter(samples_pairs)
    for tuple_outcome, amount_of_ticks in counting.items():
        global_output_state = ['9'] * number_of_qubits
        for clust_index in range(len(clusters_tuple)):
            cluster_local = clusters_tuple[clust_index]
            for qubit_index in range(len(cluster_local)):
                global_output_state[cluster_local[qubit_index]] = \
                    tuple_outcome[clust_index][1][qubit_index]

        global_output_state = ''.join(global_output_state)
        if global_output_state not in counts_dictionary_now.keys():
            counts_dictionary_now[global_output_state] = defaultdict(float)

        counts_dictionary_now[global_output_state][
            tuple([x[0] for x in tuple_outcome])] += amount_of_ticks

    return counts_dictionary_now

# def sample_from_multiple_input_states_local(input_states_dictionary):



def sample_from_multiple_input_states_alternative_broken(input_states_dictionary: Dict[str, int],
                                                  local_stochastic_matrices,
                                                  print_progres_bar=False,
                                                  old_sampling=False):
    items_range = input_states_dictionary.items()
    clusters_list = list(local_stochastic_matrices.keys())
    clusters_tuple = tuple(clusters_list)

    number_of_clusters = len(clusters_list)
    number_of_qubits = sum([len(cluster) for cluster in clusters_list])


    local_columns_big_matrices = {input_bitstring: {
        cluster: stochastic_map[:, int("".join([input_bitstring[x] for x in cluster]), 2)] for
        cluster, stochastic_map in
        local_stochastic_matrices.items()} for input_bitstring in input_states_dictionary.keys()}

    if old_sampling:
        all_samples = {cluster: [] for cluster in clusters_list}
    counts_dictionary_now = {}


    rng = np.random.default_rng()

    # all_samples = []
    for input_bitstring, number_of_samples in  tqdm(items_range,disable= not print_progres_bar):
        local_columns = local_columns_big_matrices[input_bitstring]

        samples_pairs = []
        for cluster in clusters_list:
            local_distro_now = local_columns[cluster]
            local_output_states = rng.multinomial(n=number_of_samples,
                                                  pvals=local_distro_now)

            local_samples = []
            for further_index in range(len(local_output_states)):
                local_samples += [local_registers_dictionary[len(cluster)][further_index]] * local_output_states[further_index]

            rng.shuffle(local_samples)
            # print(local_samples)

            if old_sampling:
                all_samples[cluster] += local_samples
            else:
                samples_pairs.append(local_samples)
        # all_samples.append(samples_pairs)




        if old_sampling:
            counts_dict = defaultdict(int)
            for i in range(sum(list(input_states_dictionary.values()))):
                counts_dict[tuple([all_samples[cluster][i] for cluster in clusters_list])] += 1
            # t1=time.time()

            formatted_counts_dict = {}
            for global_output_state, number_of_counts in counts_dict.items():
                formatting_state = [None for _ in range(number_of_qubits)]

                for cluster_index in range(number_of_clusters):
                    cluster_now = clusters_list[cluster_index]
                    cluster_state_bitstring = convert.integer_to_bitstring(global_output_state[cluster_index],
                                                                       len(cluster_now))

                    for qi in range(len(cluster_now)):
                        formatting_state[cluster_now[qi]] = cluster_state_bitstring[qi]
        else:
            # print(samples_pairs[0:5])
            print(samples_pairs)
            samples_pairs = list(zip(*samples_pairs))

            print(samples_pairs[0])
            print(len(samples_pairs[0]))

            # raise KeyboardInterrupt
            # print(samples_pairs[0])
            # print(samples_pairs[1])
            counting = Counter(samples_pairs)
            for tuple_outcome, amount_of_ticks in counting.items():
                # print(tuple_outcome)
                global_output_state = ['9'] * number_of_qubits
                for clust_index in range(len(clusters_tuple)):
                    cluster_local = clusters_tuple[clust_index]
                    for qubit_index in range(len(cluster_local)):
                        global_output_state[cluster_local[qubit_index]] = tuple_outcome[clust_index][1][qubit_index]
                            # tuple_outcome[clust_index][1][qubit_index]


                global_output_state = ''.join(global_output_state)
                if global_output_state not in counts_dictionary_now.keys():
                    counts_dictionary_now[global_output_state] = defaultdict(float)

                counts_dictionary_now[global_output_state][
                    tuple([x[0] for x in tuple_outcome])] += amount_of_ticks

    formatted_counts_dict = dict(counts_dictionary_now)
    return formatted_counts_dict


def sample_from_multiple_input_states_alternative(input_states_dictionary: Dict[str, int],
                                                  local_stochastic_matrices,
                                                  print_progres_bar=False,
                                                  old_sampling=False,):
    items_range = input_states_dictionary.items()

    clusters_tuple = tuple(list(local_stochastic_matrices.keys()))

    if print_progres_bar:
        items_range = tqdm(items_range)

    local_columns_big_matrices = {input_bitstring: {
        cluster: stochastic_map[:, int("".join([input_bitstring[x] for x in cluster]), 2)] for
        cluster, stochastic_map in
        local_stochastic_matrices.items()} for input_bitstring in input_states_dictionary.keys()}

    clusters_list = list(local_stochastic_matrices.keys())
    number_of_clusters = len(clusters_list)
    number_of_qubits = sum([len(cluster) for cluster in clusters_list])

    all_samples = {cluster: [] for cluster in clusters_list}
    counts_dictionary_now = {}

    local_registers_dictionary = {len(cluster):
                                      [tuple(convert.integer_to_bitstring(index_outcome, len(cluster)))
                                       for index_outcome in range(int(2 ** len(cluster)))]
                                  for cluster in clusters_list}


    rng = np.random.default_rng()
    counts_dictionary_now = {}
    import time
    t0 = time.time()

    counting_all = {}
    for input_bitstring, number_of_samples in items_range:
        local_columns = local_columns_big_matrices[input_bitstring]

        samples_pairs = []
        for cluster in clusters_list:
            local_distro_now = local_columns[cluster]
            local_output_states = rng.multinomial(n=number_of_samples,
                                                  pvals=local_distro_now)

            local_samples = []
            for further_index in range(len(local_output_states)):
                local_samples += [local_registers_dictionary[len(cluster)][further_index]] * local_output_states[further_index]

            rng.shuffle(local_samples)
            # print(local_samples)

            if old_sampling:
                all_samples[cluster] += local_samples
            else:
                samples_pairs.append(local_samples)
            # print(len(samples_pairs))
        # print(samples_pairs[0])

        if old_sampling:
            counts_dict = defaultdict(int)
            for i in range(sum(list(input_states_dictionary.values()))):
                counts_dict[tuple([all_samples[cluster][i] for cluster in clusters_list])] += 1
            # t1=time.time()

            formatted_counts_dict = {}
            for global_output_state, number_of_counts in counts_dict.items():
                formatting_state = [None for _ in range(number_of_qubits)]

                for cluster_index in range(number_of_clusters):
                    cluster_now = clusters_list[cluster_index]
                    cluster_state_bitstring = convert.integer_to_bitstring(global_output_state[cluster_index],
                                                                       len(cluster_now))

                    for qi in range(len(cluster_now)):
                        formatting_state[cluster_now[qi]] = cluster_state_bitstring[qi]
        else:
            # print(samp    les_pairs[0])
            # print(clusters_list)
            # print(number_of_clusters)
            # print(samples_pairs[0][0])
            # print([samples_pairs[k][0] for k in range(number_of_clusters)])
            samples_pairs = list(zip(*samples_pairs))
            # print(len(samples_pairs))
            #
            # print(samples_pairs[0])


            counting = Counter(samples_pairs)
            for tuple_outcome, amount_of_ticks in counting.items():
                # print(tuple_outcome)
                global_output_state = ['9'] * number_of_qubits
                for clust_index in range(len(clusters_tuple)):
                    cluster_local = clusters_tuple[clust_index]
                    for qubit_index in range(len(cluster_local)):
                        global_output_state[cluster_local[qubit_index]] = \
                            tuple_outcome[clust_index][qubit_index]

                global_output_state = ''.join(global_output_state)
                if global_output_state not in counts_dictionary_now.keys():
                    counts_dictionary_now[global_output_state] = 0

                counts_dictionary_now[global_output_state] += amount_of_ticks

    t1= time.time()
    qprint("This took:",t1-t0)
    formatted_counts_dict = dict(counts_dictionary_now)
    return formatted_counts_dict

def _get_clusters_samples(counts_dictionary,
                            local_stochastic_matrices,
                            local_columns_big_matrices,
                            local_registers_dictionary
                          ):
    rng = np.random.default_rng()
    items_range = counts_dictionary.items()

    clusters_list = list(local_stochastic_matrices.keys())

    counting_all = []
    for input_bitstring, number_of_samples in items_range:
        local_columns = local_columns_big_matrices[input_bitstring]

        samples_pairs = []
        for cluster in clusters_list:
            local_distro_now = local_columns[cluster]
            local_output_states = rng.multinomial(n=number_of_samples,
                                                  pvals=local_distro_now)

            local_samples = []
            for further_index in range(len(local_output_states)):
                local_samples += [local_registers_dictionary[len(cluster)][further_index]] * \
                                 local_output_states[further_index]

            rng.shuffle(local_samples)

            samples_pairs.append(local_samples)

        samples_pairs = list(zip(*samples_pairs))

        counting_all = counting_all + samples_pairs
    return counting_all

import time
def _get_clusters_samples_mp(items_range,
                             other_stuff
                          ):
    local_stochastic_matrices = other_stuff['local_stochastic_matrices']
    local_columns_big_matrices = other_stuff['local_columns_big_matrices']
    local_registers_dictionary = other_stuff['local_registers_dictionary']


    rng = np.random.default_rng()

    clusters_list = list(local_stochastic_matrices.keys())

    counting_all = []
    # counting_all = {}
    for input_bitstring, number_of_samples in items_range:
        local_columns = local_columns_big_matrices[input_bitstring]

        samples_pairs = []
        for cluster in clusters_list:
            local_distro_now = local_columns[cluster]
            local_output_states = rng.multinomial(n=number_of_samples,
                                                  pvals=local_distro_now)

            local_samples = []
            for further_index in range(len(local_output_states)):
                local_samples += [local_registers_dictionary[len(cluster)][further_index]] * \
                                 local_output_states[further_index]

            rng.shuffle(local_samples)

            samples_pairs.append(local_samples)

        samples_pairs = list(zip(*samples_pairs))

        # counting_all = counting_all + samples_pairs

        counting_all = counting_all+samples_pairs

    return {f"{time.time()}, {np.random.randint(0,1000)}":counting_all}


def _count_batches(counting_all,
                   number_of_batches,
                   number_of_qubits,
                   clusters_tuple):


    total_length = len(counting_all)
    batches_size = int(total_length//number_of_batches)

    counts_dicts = []
    for batch_index in range(number_of_batches):
        if batch_index<number_of_batches-1:
            counts_now = counting_all[batch_index*batches_size:(batch_index+1)*batches_size]
        else:
            counts_now = counting_all[batch_index * batches_size:]

        counts_dictionary_now = {}
        counting = Counter(counts_now)
        for tuple_outcome, amount_of_ticks in counting.items():
            global_output_state = ['9'] * number_of_qubits
            for clust_index in range(len(clusters_tuple)):
                cluster_local = clusters_tuple[clust_index]
                for qubit_index in range(len(cluster_local)):
                    global_output_state[cluster_local[qubit_index]] = \
                        tuple_outcome[clust_index][qubit_index]
            global_output_state = ''.join(global_output_state)
            if global_output_state not in counts_dictionary_now.keys():
                counts_dictionary_now[global_output_state] = 0

            counts_dictionary_now[global_output_state] += amount_of_ticks
        counts_dicts.append(counts_dictionary_now)
    return counts_dicts


def _count_batches_mp(batches_tuples,
                   other_stuff):
    # number_of_batches = other_stuff['number_of_batches']
    number_of_qubits = other_stuff['number_of_qubits']
    clusters_tuple = other_stuff['clusters_tuple']

    counts_dicts = []
    for batch_of_counts in batches_tuples:
        for counts_now in batch_of_counts:
            counts_dictionary_now = {}
            counting = Counter(counts_now)
            for tuple_outcome, amount_of_ticks in counting.items():
                global_output_state = ['9'] * number_of_qubits
                for clust_index in range(len(clusters_tuple)):
                    cluster_local = clusters_tuple[clust_index]
                    for qubit_index in range(len(cluster_local)):
                        global_output_state[cluster_local[qubit_index]] = \
                            tuple_outcome[clust_index][qubit_index]
                global_output_state = ''.join(global_output_state)
                if global_output_state not in counts_dictionary_now.keys():
                    counts_dictionary_now[global_output_state] = 0

                counts_dictionary_now[global_output_state] += amount_of_ticks
            counts_dicts.append(counts_dictionary_now)

    return {f"{time.time()}, {np.random.randint(0,100)}":counts_dicts}


def sample_from_multiple_input_states_alternative_batches(input_states_dictionary: Dict[str, int],
                                                  local_stochastic_matrices,
                                                          number_of_batches,
                                                  print_progres_bar=False,
                                                  old_sampling=False,
                                                          multiprocessing=False):
    items_range = input_states_dictionary.items()
    clusters_list = list(local_stochastic_matrices.keys())
    clusters_tuple = tuple(list(local_stochastic_matrices.keys()))

    # if print_progres_bar:
    #     items_range = tqdm(items_range)

    local_columns_big_matrices = {input_bitstring: {
        cluster: stochastic_map[:, int("".join([input_bitstring[x] for x in cluster]), 2)] for
        cluster, stochastic_map in
        local_stochastic_matrices.items()} for input_bitstring in input_states_dictionary.keys()}


    # number_of_clusters = len(clusters_list)
    number_of_qubits = sum([len(cluster) for cluster in clusters_list])

    # all_samples = {cluster: [] for cluster in clusters_list}
    # counts_dictionary_now = {}

    local_registers_dictionary = {len(cluster):
                                      [tuple(convert.integer_to_bitstring(index_outcome, len(cluster)))
                                       for index_outcome in range(int(2 ** len(cluster)))]
                                  for cluster in clusters_list}

    rng = np.random.default_rng()
    # counts_dictionary_now = defaultdict(float)
    import time
    t0 = time.time()

    # if multiprocessing:

    t0 = time.time()
    if multiprocessing:

        main_tuples = list(items_range)

        res_mp = utils.wrapped_multiprocessing_function(tuple_of_main_arguments=main_tuples,
                                                        additional_kwargs={'local_stochastic_matrices':local_stochastic_matrices,
                                                                           'local_registers_dictionary':local_registers_dictionary,
                                                                           'local_columns_big_matrices':local_columns_big_matrices},
                                                        function_to_multiprocess=_get_clusters_samples_mp,
                                                        printing=True)
        print(time.time()-t0)
        counting_all = []
        for value in res_mp.values():
            counting_all = counting_all+value
        # print(time.time()-t0)
        #
        print(len(list(res_mp.values())))



    else:
        counting_all = _get_clusters_samples(counts_dictionary=input_states_dictionary,
                                             local_stochastic_matrices=local_stochastic_matrices,
                                             local_registers_dictionary=local_registers_dictionary,
                                             local_columns_big_matrices=local_columns_big_matrices)
    rng.shuffle(counting_all)
    rng.shuffle(counting_all)
    rng.shuffle(counting_all)
    t1 = time.time()

    if multiprocessing:
        total_length = len(counting_all)
        batches_size = int(total_length // number_of_batches)

        batches_mp = []
        for batch_index in range(number_of_batches):
            if batch_index < number_of_batches - 1:
                counts_now = counting_all[batch_index * batches_size:(batch_index + 1) * batches_size]
            else:
                counts_now = counting_all[batch_index * batches_size:]

            batches_mp.append((counts_now,))


        res_mp = utils.wrapped_multiprocessing_function(tuple_of_main_arguments=batches_mp,
                                                        additional_kwargs={'number_of_qubits':number_of_qubits,
                                                                           'clusters_tuple':clusters_tuple,
                                                                           },
                                                        function_to_multiprocess=_count_batches_mp,
                                                        printing=True)

        # print(len(list(res_mp.values())))

        counts_dicts = []
        for value in res_mp.values():
            counts_dicts = counts_dicts+value

    else:
        counts_dicts = _count_batches(counting_all,
                       number_of_batches,
                       number_of_qubits,
                       clusters_tuple)

    t2 = time.time()
    qprint("Sampling took:", t1 - t0)
    qprint("Counting took:", t2 - t1)
    # formatted_counts_dict = dict(counts_dictionary_now)
    return counts_dicts
           # normalizations


def sample_from_product_noise_model(input_bitstring: str,
                                    local_noise_matrices: Dict[Tuple[int], np.ndarray],
                                    number_of_samples: int,
                                    local_columns_big_dictionary=None):
    clusters_list = list(local_noise_matrices.keys())
    number_of_clusters = len(clusters_list)
    number_of_qubits = len(input_bitstring)

    if local_columns_big_dictionary is not None:
        if not isinstance(input_bitstring,str):
            input_bitstring = ''.join(input_bitstring)
            # print(local_columns_big_dictionary)
        local_columns = local_columns_big_dictionary[input_bitstring]
    else:
        local_columns = {
            cluster: stochastic_map[:, int("".join([input_bitstring[x] for x in cluster]), 2)]
            for cluster, stochastic_map in local_noise_matrices.items()}

    new_sampling = True

    counts_dict = defaultdict(int)

    if new_sampling:
        all_samples = {}
        # t0 = time.time()
        rng = np.random.default_rng()
        for cluster in clusters_list:
            # t2 =time.time()
            local_distro_now = local_columns[cluster]
            # rng = np.random.default_rng()
            local_output_states = rng.multinomial(n=number_of_samples,
                                                  pvals=local_distro_now)

            local_samples = []
            for index_outcome in range(len(local_output_states)):
                local_samples += [convert.integer_to_bitstring(index_outcome, len(cluster))] * \
                                 local_output_states[index_outcome]

            #         list(np.full(shape=local_output_states[index_outcome],
            #                                   fill_value=convert.integer_to_bitstring(index_outcome,
            #                                                                       len(cluster))
            #                                   ))
            # # t3 = time.time()

            rng.shuffle(local_samples)
            all_samples[cluster] = local_samples

        # t1 =time.time()
        # t5 = time.time()
        for i in range(number_of_samples):
            global_output_state = [None for __ in range(number_of_qubits)]
            for cluster in clusters_list:
                local_output_state = all_samples[cluster][i]
                for qubit_index in range(len(cluster)):
                    global_output_state[cluster[qubit_index]] = local_output_state[qubit_index]

            global_output_state = ''.join(global_output_state)
            counts_dict[global_output_state] += 1

        counts_dict_formatted = counts_dict
        # t6 = time.time()
        # if (t6-t5)>(t1 - t0):
        #     print('sampling:', t1 - t0)
        #     print('formatting:', t6-t5)
    else:

        for _ in range(number_of_samples):
            output_now = np.zeros(number_of_clusters, dtype=tuple)
            for cluster_index in range(number_of_clusters):
                cluster = clusters_list[cluster_index]
                local_distro = local_columns[cluster]
                local_sample_now = np.random.multinomial(pvals=local_distro,
                                                         n=1)
                output_now[cluster_index] = tuple(local_sample_now)
            counts_dict[tuple(output_now)] += 1

        counts_dict_formatted = __format_counts_from_clusters(counts_dict=counts_dict,
                                                              clusters_list=clusters_list)

    # for cluster_index in range(number_of_clusters):
    #     cluster = clusters_list[cluster_index]
    #     local_distro = local_columns[cluster]
    #     local_sample_now = np.random.multinomial(pvals=local_distro,
    #                                              n=number_of_samples)
    #
    #
    #
    #
    #

    return counts_dict_formatted
#
# if __name__ == 'main':
#
#     local_noise_matrices_base = {
#         (0, 1, 7): array([[7.57029412e-01, 1.74052381e-01, 1.01661290e-01, 2.20660714e-02,
#                            1.97447458e-01, 4.16703125e-02, 2.80763636e-02, 5.30517241e-03],
#                           [1.76901961e-01, 5.75815873e-01, 2.31080645e-02, 4.25196429e-02,
#                            1.79110169e-01, 1.54307813e-01, 1.62727273e-02, 1.00879310e-02],
#                           [1.86215686e-02, 1.22015873e-02, 7.29356452e-01, 3.43464286e-01,
#                            6.38983051e-03, 2.61562500e-03, 2.53301818e-01, 7.51551724e-02],
#                           [1.51764706e-03, 1.30714286e-02, 1.02304839e-01, 4.27332143e-01,
#                            3.04745763e-03, 3.31718750e-03, 1.08665455e-01, 9.43500000e-02],
#                           [3.59137255e-02, 9.53571429e-02, 4.74838710e-03, 8.10535714e-03,
#                            4.08279661e-01, 1.30548438e-01, 5.86600000e-02, 1.96810345e-02],
#                           [9.04901961e-03, 1.23300000e-01, 1.08709677e-03, 6.53214286e-03,
#                            1.92888136e-01, 6.40910937e-01, 2.06563636e-02, 5.24793103e-02],
#                           [8.86274510e-04, 4.48412698e-03, 3.30709677e-02, 9.45196429e-02,
#                            1.07423729e-02, 1.02140625e-02, 4.17769091e-01, 2.78255172e-01],
#                           [8.03921569e-05, 1.71746032e-03, 4.66290323e-03, 5.54607143e-02,
#                            2.09491525e-03, 1.64156250e-02, 9.65981818e-02, 4.64686207e-01]]),
#         (2,): array([[0.96430696, 0.08032241],
#                      [0.03569304, 0.91967759]]), (3,): array([[0.91271138, 0.03638376],
#                                                               [0.08728862, 0.96361624]]),
#         (4,): array([[0.9692458, 0.05537525],
#                      [0.0307542, 0.94462475]]), (5,): array([[0.92232226, 0.77049618],
#                                                              [0.07767774, 0.22950382]]),
#         (6,): array([[0.95776162, 0.04182947],
#                      [0.04223838, 0.95817053]]), (8,): array([[0.97306899, 0.1251291],
#                                                               [0.02693101, 0.8748709]]),
#         (9,): array([[0.93448873, 0.07061062],
#                      [0.06551127, 0.92938938]]), (10,): array([[0.98673934, 0.37029401],
#                                                                [0.01326066, 0.62970599]]),
#         (11,): array([[0.97976416, 0.0407595],
#                       [0.02023584, 0.9592405]]), (12,): array([[0.9721489, 0.04683174],
#                                                                [0.0278511, 0.95316826]]),
#         (13,): array([[0.95634948, 0.0379854],
#                       [0.04365052, 0.9620146]]), (14,): array([[0.98572483, 0.04384909],
#                                                                [0.01427517, 0.95615091]]),
#         (16,): array([[0.9194975, 0.06026644],
#                       [0.0805025, 0.93973356]]),
#         (15, 17): array([[9.47176812e-01, 3.60669643e-02, 4.42798387e-02, 8.97796610e-03],
#                          [2.44862319e-02, 9.33566964e-01, 2.00887097e-03, 3.23272881e-01],
#                          [2.74797101e-02, 1.12142857e-03, 9.26055645e-01, 3.32940678e-02],
#                          [8.57246377e-04, 2.92446429e-02, 2.76556452e-02, 6.34455085e-01]]),
#         (18,): array([[0.98574926, 0.05787736],
#                       [0.01425074, 0.94212264]]), (19,): array([[0.96944702, 0.07846338],
#                                                                 [0.03055298, 0.92153662]])}
#
#     p, q, p2, q2 = 0.2, 0.3, 0.1, 0.15
#
#     p3, q3 = 0.05, 0.09
#     # local_noise_matrices_base = {(0,): np.array([[1 - p, q],
#     #                                              [p, 1 - q]]),
#     #                              (1,): np.array([[1 - p2, q2],
#     #                                              [p2, 1 - q2]]),
#     #                              (2,): np.array([[1 - p3, q3],
#     #                                              [p3, 1 - q3]]),
#     #                              }
#     number_of_samples = 10 ** 5
#     number_of_qubits = 10
#     dimension = int(2 ** number_of_qubits)
#
#     local_noise_matrices = {}
#     for cluster in local_noise_matrices_base.keys():
#         if np.any(np.array(cluster) >= number_of_qubits):
#             break
#         else:
#             local_noise_matrices[cluster] = local_noise_matrices_base[cluster].real
#
#     input_bitstring = []
#     for i in range(number_of_qubits):
#         if i % 2 == 0:
#             input_bitstring.append("0")
#         else:
#             input_bitstring.append("0")
#
#     input_bitstring = "".join(input_bitstring)
#
#     if number_of_qubits <= 10:
#         global_noise_matrix = np.eye(dimension, dtype=float)
#         for cluster, local_noise_matrix in local_noise_matrices.items():
#             print(cluster)
#             global_noise_matrix = global_noise_matrix @ quanf.embed_operator_in_bigger_hilbert_space(
#                 number_of_qubits=number_of_qubits,
#                 local_operator=local_noise_matrix,
#                 global_indices=cluster).real
#
#         true_distro = global_noise_matrix[:, int(input_bitstring, 2)]
#
#
#     t0 = time.time()
#     counts_dict_formatted_first = sample_from_product_noise_model(input_bitstring=input_bitstring,
#                                     local_noise_matrices=local_noise_matrices,
#                                     number_of_samples=number_of_samples)
#
#     t1 = time.time()
#
#
#     clusters_list = list(local_noise_matrices.keys())
#     number_of_clusters = len(clusters_list)
#
#     local_columns = {cluster: stochastic_map[:, int("".join([input_bitstring[x] for x in cluster]), 2)]
#                      for cluster, stochastic_map in local_noise_matrices.items()}
#
#     counts_dict = defaultdict(float)
#     cluster_samples_original = {}
#     for cluster_index in tqdm(range(number_of_clusters)):
#         cluster = clusters_list[cluster_index]
#         local_distro = local_columns[cluster]
#         local_samples_now = np.random.multinomial(pvals=local_distro,
#                                                   n=number_of_samples)
#         cluster_samples_original[cluster] = local_samples_now
#
#     from povms_qi import ancillary_functions as anf_pt
#
#
#     # samples_clusters_shuffled = {}
#
#     samples_list_shuffled = []
#     for cluster, local_samples in tqdm(cluster_samples_original.items()):
#         cluster_samples_local = []
#         for index in range(len(local_samples)):
#             cluster_samples_local = cluster_samples_local + list(np.full(shape=local_samples[index],
#                                                                          fill_value=convert.integer_to_bitstring(
#                                                                              integer=index,
#                                                                              number_of_bits=len(cluster))))
#         local_samples_shuffled = anf_pt.properly_shuffle_stuff(cluster_samples_local)
#         samples_list_shuffled.append(local_samples_shuffled)
#
#     samples_list_shuffled = np.array(samples_list_shuffled)
#
#     counts_dict = defaultdict(float)
#     for sample_index in range(number_of_samples):
#         output_now = np.zeros(number_of_clusters, dtype=tuple)
#         for cluster_index in range(number_of_clusters):
#             cluster = clusters_list[cluster_index]
#             local_sample_now = samples_list_shuffled[cluster_index,sample_index]
#             output_now[cluster_index] = tuple(local_sample_now)
#         counts_dict[tuple(output_now)] += 1
#
#     counts_dict_formatted_second = __format_counts_from_clusters(counts_dict=counts_dict,
#                                                           clusters_list=clusters_list)
#
#     t2 =time.time()
#
#
#     print(t1-t0,t2-t1)
#
#     # counts_dict_formatted_first
#
#
#     if number_of_qubits <= 10:
#         sampled_direct = np.array(np.random.multinomial(pvals=true_distro / sum(true_distro),
#                                                         n=number_of_samples)) / number_of_samples
#
#         estimated_distro_first = fda.convert_counts_dictionary_to_probability_distribution(counts_dictionary=counts_dict_formatted_first)
#         estimated_distro_second = fda.convert_counts_dictionary_to_probability_distribution(
#             counts_dictionary=counts_dict_formatted_second)
#
#         print(1/2*np.linalg.norm(true_distro-sampled_direct))
#         print(1 / 2 * np.linalg.norm(true_distro - estimated_distro_first))
#         print(1 / 2 * np.linalg.norm(true_distro - estimated_distro_second))
#
#
#



"""qrem.types.circuit_collection module contains CircuitCollection class, which serves as the base class that contains all information
about set of circuits to be sent to quantum machine for characterisation/mitigation/benchmarking purposes.
It inherits from DataStructureBase, and is extended by ExperimentResults - class that contains results of experiments performed on a given circuit collection.
""" 

from datetime import datetime
from qrem.types.datastructure_base import DataStructureBase
from qrem.common.config import QremConfigLoader
from qrem.common.constants import CIRCUIT_DATA_TYPE as c_type

import numpy as np
import numpy.typing as npt

# TODO sort out repeating circuits already in CircuitCollection

class CircuitCollection(DataStructureBase):
    """
    The class contains data about circuits to be prepared on a certain device, for a specified experiment.
    Parameters
    ----------
    device (str): name of the device on which the circuits will be run
    experiment_type (str): DDOT or QDOT (Diagonal/Quantum Detector Overlapping Tomography)
    circuits_list (list[str]): list of circuits, where each circuit is a string of symbols (0-1 for DDOT, 0-5 for QDOT)
    qubit_indices (list[int]): list of sorted indices of qubits on which the circuits are to be executed
    gate_error_threshold (float): gate error on qubit above which qubits are disregarded (we don't execute circuits there)
    locality (int): what degree marginals we wish to characterize
    no_shots (int): how many times each circuit is executed
    datetime_created_utc: when the circuit collection was created
    author (str): creator of collection
    """
    def __init__(self, device):
        # [X] default value definitions
        super().__init__()
        self.device = device
        self.experiment_type = ''
        
        self.circuits_labels = []           # it is good to keep circuits still as a list of strings for results analysis
        self.circuits = []                  # target format: np.ndarray((number_of_qubits,number_of_circuits),dtype=uint8)
        #self.circuits_dictionary = []

        self.qubit_indices = []             #needs to be sorted - valid qubits indices
        self.random_circuits_count = None #to be filled in compute_number_of_circuits()
        self.total_circuits_count = None    #to be filled in compute_number_of_circuits(); dividing it by random_circuits_count will give how many times a circuit needs to be run

        self.gate_error_threshold = None  # 0.01
        self.locality = None  # 2
        self.no_shots = None  # 10000
        self.datetime_created_utc = datetime.utcnow()
        self.experiment_name = ''
        self.author = ''
        self.metadata = []

    def load_config(self, config: QremConfigLoader):
        self.experiment_name = config.experiment_name
        self.experiment_type = config.experiment_type
        self.locality = config.k_locality
        self.gate_error_threshold = config.gate_threshold
        self.author = config.author
        self.no_shots = config.shots_per_circuit

    def load_from_dict(self, dictionary):
        for key in dictionary:
            if key in self.get_dict_format():
                setattr(self, key,
                        dictionary[key])  # TODO: secure against bad data, e.g mismatched qubit and circuit len?
                
                if key =="circuits":
                    print(type(dictionary[key]) is list)
                    if type(dictionary[key]) is list:
                        self.circuits = np.array(dictionary[key],dtype=c_type)
                    elif type(dictionary[key]) is  np.ndarray:
                        if(type(dictionary[key][0][0])!=c_type):
                            print("WARNING, Type of ndarray element should be uint8, converting")
                            self.circuits = self.circuits.astype(c_type)
                        pass
                    else:
                        print("WARNIGN, Werid format of circuits saved in json, may fail on next step")
                        self.circuits = np.ndarray(dictionary[key],dtype=c_type)
        pass



# tests:
if __name__ == "__main__":
    test_collection = CircuitCollection('test_name')
    print(test_collection.get_dict_format())
    dictionary_to_load = {'experiment_type': 'qdot',
                          'circuits': np.array([[1, 2, 3], [4, 5, 6]], dtype=c_type),
                          'qubit_indices': [0, 2, 5],
                          'gate_error_threshold': 0.005,
                          'no_shots': 1,
                          'datetime_created_utc': datetime.utcnow(),
                          'author': 'tester',
                          'metadata': {"note1": 'some string note',}}
    test_collection.load_from_dict(dictionary_to_load)
    print(test_collection.circuits.shape)
    print('after loading dict:\n', test_collection.get_dict_format())
    json_dict_test = test_collection.to_json()
    test_collection.export_json('test_exported_json.json', overwrite=True)
    test_collection.import_json('test_exported_json.json')
    print('after reading from file:\n', test_collection.get_dict_format())
    # pickle_dict_test = test_collection.get_pickle()
    # test_collection.export_pickle('exported_pickle', overwrite=True)
    # test_collection.metadata["note2"]= 'changed note'
    # print('after change:\n', test_collection.get_dict_format())
    # test_collection.import_pickle('exported_pickle')



"""qrem.common.utils module contains all helpful functions used throughout the projcets.

Current contents of this module will relate to:
- dit/bitstring conversions
- python/numpy format conversions
- boolean operators on lists (that treat lists as sets)
- other helpful functions not covered elswhere

May be split into other modules in the future (very likely)

"""
import sys
import inspect
import warnings
import multiprocessing
from typing import List, Dict, Iterable, Optional, Callable

import numpy as np

from qrem.common.printer import qprint



#===================================================
# Boolean operation on lists, dicts and  sets
#===================================================

def lists_intersection(lst1: list,
                       lst2: list):
    """Intersection of 2 lists lst1 and lst2 after conversion of both to sets.
    
    Returns
    -------
    list

    Notes
    -----
    Multiple entries within each of the lists will merge 
    """
    return list(set(lst1) & set(lst2))

def lists_difference(lst1: list,
                     lst2: list):
    """Difference between 2 lists lst1 and lst2 after conversion of both to sets.
    
    Returns
    -------
    list
    
    Notes
    -----
    Multiple entries within each of the lists will merge 
    """
    return list(set(lst1) - set(lst2))

def lists_sum(lst1: list,
              lst2: list):
    """Union between 2 lists lst1 and lst2 after conversion of both to sets. 

    Returns
    -------
    list
    
    Notes
    -----
    Multiple entries within each of the lists will merge 
    """
    return list(set(lst1).union(set(lst2)))

def lists_sum_multiple(lists: List[list]):
    """Union between all lists in the input  list of lists (lists) after conversion to sets. 
    
    Returns
    -------
    list

    Notes
    -----
    Multiple entries within each of the lists will merge 
    """
    return list(set().union(*lists))

def lists_intersection_multiple(lists: List[list]):
    """Intersection of all lists in the input  list of lists (lists) after conversion  to sets. Returns a list.

    Notes
    -----
    Multiple entries within each of the lists will merge 
    """
    l0 = lists[0]
    l1 = lists[1]

    int_list = lists_intersection(l0, l1)
    for l in lists[2:]:
        int_list = lists_intersection(int_list, l)
    return int_list

#===================================================
# Checks and operations on lists, dicts etc
#===================================================
#TEST_ME test and improve efficiency of the check
def check_for_multiple_occurences(lists: List[list]) -> bool:    
    """Checks for duplicate elements between multiple lists. Multiple occurences within each of lists are not taken into consideration

    Returns
    -------
    bool
    """
    # possible reimplementation, not sure if faster / slower
    # u = reduce(set.union, map(set, ll))
    # sd = reduce(set.symmetric_difference, map(set, lists))
    # len(u - sd) != 0 ->there are repeating elements
    for i in range(len(lists)):
        for j in range(i + 1, len(lists)):
            if len(lists_intersection(lists[i], lists[j])) != 0:
                return True

    return False

def enumerate_dict(list: List[int])->Dict:
    ''' This function takes in a list 'list' and returns a dictionary where the keys are the indices of the elements in 'some_list' 
    and the values are the elements themselves.

    Parameters
    ----------
    list: List[int]
        list to be enumerated
    '''
    return dict(enumerate(list))

def sort_things(stuff_to_sort: Iterable, according_to_what: Iterable) -> Iterable:
    """Sort stuff according to some other stuff assuming that the stuff_to_sort 
    is sorted in natural order (0, 1, 2, ...)"""
    return [element for _, element in sorted(zip(according_to_what, stuff_to_sort), key=lambda pair: pair[0])]

def swap_keys_and_values(enumerated_dict: Dict[int, int]) -> Dict[int, int]:
    """
   This function takes in a dictionary 'enumerate' where the keys are integers and the values are also integers. 
   It returns a new dictionary where the keys and values are reversed from the input dictionary

    Parameters
    ----------
    enumerated_dict: Dict[int, int]

    """
    reversed_map = {}
    for index_sorted, true_index in enumerated_dict.items():
        if true_index in reversed_map:
            raise warnings.warn(f"# WARNING # Value in the list repeats itself. Would silently merge and discard this element. Maybe should be like that?")
        reversed_map[true_index] = index_sorted
    return reversed_map

# formerly get_reversed_enumerated_from_indices
def map_index_to_order(indices: List[int]) -> Dict[str, int]:
    """
    Given indices list, enumerate them and return map which is inverse of enumerate
    Parameters
    ----------
    indices: List[int]
        list of indices (e.g. qubit indices), that will be mapped in a dict to their natural order 
    """
    return swap_keys_and_values(enumerate_dict(indices))

#===================================================
# Helpful, programming funtionts
#===================================================

def wrapped_multiprocessing_function(tuple_of_main_arguments: List[tuple],
                                     additional_kwargs: dict,
                                     function_to_multiprocess: Callable,
                                     number_of_threads: Optional[int] = None,
                                     # debug=False,
                                     printing=False):
    """Wrapper for executing a custom defined function using multiprocessing over a list of variable arguments. 
    Additional list of constant arguments can be provided 

    Parameters
    ----------
    tuple_of_main_arguments: List[tuple]
        This is list of tuples that is divided into batches that are passed to different threads of
        multiprocessing. This, therefore should be set of VARIABLE arguments that are passed to the
        function of interest

    additional_kwargs: dict
        This is dictionary of arguments that are CONSTANT for all function evaluations.
        The dictionary is COPIED for each thread and passed to the function.

    function_to_multiprocess: Callable
    number_of_threads: int
    """

    if number_of_threads is None:
        number_of_threads = int(
            np.min([multiprocessing.cpu_count() - 1, len(tuple_of_main_arguments)]))

    length = len(tuple_of_main_arguments)
    # print(length)
    # raise KeyboardInterrupt
    # IF there is less arguments than threads, we reduce number of threads
    if length < number_of_threads:
        number_of_threads = length


    division_cores = length // number_of_threads

    all_indices, arguments = [], []
    for process_pool_index in range(number_of_threads):
        if process_pool_index == number_of_threads - 1:
            slice_now = slice(process_pool_index * division_cores,
                              -1)
        else:
            slice_now = slice(process_pool_index * division_cores,
                              (process_pool_index + 1) * division_cores)
        sample_indices_now = tuple_of_main_arguments[slice_now]
        arguments.append((sample_indices_now,
                          additional_kwargs))

    if printing:
        qprint(f'Running {number_of_threads} threads!', '', 'green')

    pool = multiprocessing.Pool(number_of_threads)
    results = pool.starmap_async(function_to_multiprocess,
                                 arguments)
    pool.close()
    pool.join()

    res_multiprocessing = results.get()

    results_dictionary_all = {}

    for dict_now in res_multiprocessing:
        results_dictionary_all = {**results_dictionary_all, **dict_now}

    return results_dictionary_all 

def get_full_object_size(obj, seen=None):
    """Recursively finds size of objects in bytes"""
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if hasattr(obj, '__dict__'):
        for cls in obj.__class__.__mro__:
            if '__dict__' in cls.__dict__:
                d = cls.__dict__['__dict__']
                if inspect.isgetsetdescriptor(d) or inspect.ismemberdescriptor(d):
                    size += get_full_object_size(obj.__dict__, seen)
                break
    if isinstance(obj, dict):
        size += sum((get_full_object_size(v, seen) for v in obj.values()))
        size += sum((get_full_object_size(k, seen) for k in obj.keys()))
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        try:
            size += sum((get_full_object_size(i, seen) for i in obj))
        except TypeError:
            print("Unable to get size of %r. This may lead to incorrect sizes. Please report this error.", obj)
    if hasattr(obj, '__slots__'): # can have __slots__ with __dict__
        size += sum(get_full_object_size(getattr(obj, s), seen) for s in obj.__slots__ if hasattr(obj, s))
        
    return size

# import numpy as np
# import orjson
# from pathlib import Path
# class ExampleDataClass:
#     def __init__(self, name):

#         #[X] default value definitions
#         self.name = name                # name of the Data set
#         dt = np.dtype(np.uint8)         # type definition - helper def for below
#         self.circuits = np.array([],dt) # some data type (for circuits it should be probably list of np.arrays or array of arrays)
#         self.id = -1                    # unique id, set manually during export (field not used currently)
#         self.experiment_type = "DDOT"   # type of experiment - can be coded as string, or as sth else
#         pass

#     def getDictFormat(self):
#         '''returns this class as a json-like dictionary structure'''
#         return self.__dict__

#     def getJSON(self):
#         return  orjson.dumps(self, default=lambda o: o.__dict__,
#         option=orjson.OPT_SERIALIZE_NUMPY, sort_keys=True)
    
#     def exportJSON(self,json_export_path,overwrite = True):
        
#         #[7] Save into json file
#         if(Path(json_export_path).is_file() and not overwrite):
#             print(f"WARNING:: Ommiting export to existing file: <{json_export_path}>")     
#         else:
#             with open(json_export_path, 'w') as outfile:
#                 outfile.write(self.getJSON())

#     def importJSON(self,json_import_path):
#         '''import JSON'''
#         #stub, easy to implement
#         #first orjson.load contents of the file
#         #then assign elements of loaded dict into the class fields
#         pass



#ORGANIZATION - dont know what is the purpose of this file (MO) - (PP): important for defining initialization for packages/subpackages in python (think of every folder as a package/subpackage, when you import it __init__.py is run)
"""**qrem** module is the main module of the whole package.
It contains all qrem code after refactoring into upcoming version 1.0.0. Currently it holds
most important data structures and common functions used in the freamework. The main subpackages are: \n\n
- **common**: contains all common functions, utilities, connection to quantum backends etc. used throughout whole qrem package\n
- **cn**: csubpackage containing cluster-neighbours approach implementation \n
- **ctmp**: Mitigation procedure as described in Bravyi et al. \n
- **pipelines**: scripts containing executions of most common characterisation/mitigation routines from data generation, sending to quantum machine down to characterization and mitigation. \n
- **types**: most important data holder classess used throughout the package \n
""" 


from importlib.metadata import version
from dotenv import load_dotenv

from qrem.common.printer import qprint
from qrem.common.config import QremConfigLoader
#-----------------------
# [1] read version from installed package and provide in __version__ variable
#-----------------------
__version__ = version("qrem")

#-----------------------
# [2] read environment variables.
#-----------------------
# Set up environment variables based on the .env file https://dev.to/jakewitcher/using-env-files-for-environment-variables-in-python-applications-55a1
load_dotenv()


#-----------------------
# [3] modules binding for top-level qrem package
#-----------------------
# add here anything that should be accesible directly from "qrem.":
# example: if you want "import qrem.core.something.else" module  be available for users as "import qrem.else"

#-----------------------
# [4] MAIN PACKAGE FUNCTIONS
#-----------------------
#Main package operational  functions
def load_config(cmd_args = None, path:str = None, verbose_log = False):
    """Loads config for running main QREM pipeline

    Parameters
    ----------
    path: str, default = None
        Path to the config file to load
    verbose_log: bool, default=False
        turn on verbose logging for more printouts with info
    """

    if cmd_args != None:
        qprint(f"- LOADING CONFIG VIA CMD ARGS: ", f"{cmd_args}")
        config = QremConfigLoader.load(cmd_args=cmd_args)
    elif path != None:
        qprint(f"- LOADING CONFIG VIA FILE PATH: ", f"{path}")
        config = QremConfigLoader.load(default_path=path)
    else:
        config = QremConfigLoader.load()
    
    if(verbose_log):
        QremConfigLoader.values()

    return config


# #Main package operational  functions
# def load_config(path:str = None, verbose_log = False):
#     """Loads config for running main QREM pipeline

#     Parameters
#     ----------
#     path: str, default = None
#         Path to the config file to load
#     verbose_log: bool, default=False
#         turn on verbose logging for more printouts with info
#     """

#     if path != None:
#         print(path)
#         config = QremConfigLoader.load(default_path=path, as_dict=False)
#     else:
#         config = QremConfigLoader.load()
    
#     if(verbose_log):
#         QremConfigLoader.values()

#     return config


"""qrem.common.printer module contains all helpful functions for logging and console printouts.
For now it contains only console printing capabilities. 
In future it will be a choice - either you will be able to log output into a specified log path, print on console or both,

Current contents of this module contain:
- pretty text print and array print for preview

"""
import copy

from colorama import Fore, Style

import numpy as np
import numpy.typing as npt
import pandas as pd

#==============================================
# Helper formatters
#==============================================
def round_matrix_for_print(matrix_to_be_rounded: npt.NDArray,
                 decimal: int=3) -> npt.NDArray:
    """
    This function rounds matrix in a nice way.
    "Nice" means that it removes funny Python artifacts such as "-0.", "0j" etc.

    Parameters
    ----------
    matrix_to_be_rounded: npt.NDArray
    decimal: int

    Returns
    -------
    npt.NDArray
    """

    data_type = type(matrix_to_be_rounded[0, 0])

    first_dimension = matrix_to_be_rounded.shape[0]
    second_dimension = np.size(matrix_to_be_rounded, axis=1)

    rounded_matrix = np.zeros((first_dimension, second_dimension), dtype=data_type)

    for first_index in range(0, first_dimension):
        for second_index in range(0, second_dimension):
            real_part = round(np.real(matrix_to_be_rounded[first_index, second_index]), decimal)

            if data_type == complex or data_type == np.complex128:
                imaginary_part = round(np.imag(matrix_to_be_rounded[first_index, second_index]),
                                       decimal)
            else:
                imaginary_part = 0

            # In the following we check whether some parts are 0 and then we leave it as 0
            # Intention here is to remove some Python artifacts such as leaving "-0" instead of 0.
            # see function's description.
            if abs(real_part) != 0 and abs(imaginary_part) != 0:
                if data_type == complex or data_type == np.complex128:
                    rounded_matrix[first_index, second_index] = real_part + 1j * imaginary_part
                else:
                    rounded_matrix[first_index, second_index] = real_part
            elif abs(real_part) == 0 and abs(imaginary_part) == 0:
                rounded_matrix[first_index, second_index] = 0
            elif abs(real_part) == 0 and abs(imaginary_part) != 0:
                if data_type == complex or data_type == np.complex128:
                    rounded_matrix[first_index, second_index] = 1j * imaginary_part
                else:
                    rounded_matrix[first_index, second_index] = 0
            elif abs(real_part) != 0 and abs(imaginary_part) == 0:
                rounded_matrix[first_index, second_index] = real_part

    return rounded_matrix

def zeros_to_dots_for_print(matrix, rounding_decimal = 3) -> npt.NDArray:
    """
    This function changes zeros into dots in printing; values are treated as 0 with rounding_decimal significant digits ( default is 3) 

    Parameters
    ----------
    matrix: npt.NDArray
    rounding_decimal: int

    Returns
    -------
    npt.NDArray
    """
    m = matrix.shape[0]
    n = matrix.shape[1]

    B = np.zeros((m, n), dtype=dict)

    for i in range(m):
        for j in range(n):
            el = matrix[i, j]
            if (abs(np.round(el, rounding_decimal)) >= 0):
                B[i, j] = el
            else:
                B[i, j] = '.'

    return B

def qprint(colored_string: str, stuff_to_print = '', color='CYAN', print_floors=False) -> None:
    """A bit fancier print function that uses colors for output. 
    Should be extended/changed to have also logging capabilities

    Parameters
    ----------
    colored_string: str
        colored_string is printed with color
    stuff_to_print: str
        stuff_to_print is printed without color, after colored_string
    color: str
        color, you can use 'red','green','blue',or 'cyan' currently
    """
    if print_floors:
        print("_________________________")

    if isinstance(color, str):
        if color.lower() == 'red':
            color = Fore.RED
        elif color.lower() == 'yellow':
            color = Fore.YELLOW
        elif color.lower() == 'green':
            color = Fore.BLUE
        elif color.lower() == 'blue':
            color = Fore.GREEN
        elif color.lower() == 'cyan':
            color = Fore.CYAN
        else:
            color = Fore.CYAN

    if isinstance(stuff_to_print, str):
        if stuff_to_print == '':
            print(color + Style.BRIGHT + str(colored_string) + Style.RESET_ALL)
        elif stuff_to_print == '\n':
            print(color + Style.BRIGHT + str(colored_string) + Style.RESET_ALL)
            print()
        else:
            print(color + Style.BRIGHT + str(colored_string) + Style.RESET_ALL, repr(stuff_to_print))
    else:
        print(color + Style.BRIGHT + str(colored_string) + Style.RESET_ALL, repr(stuff_to_print))
    if print_floors:
        print("_________________________")

def warprint(colored_string: str, stuff_to_print = '', color='YELLOW', print_floors=False) -> None:
    qprint(colored_string=colored_string , stuff_to_print = stuff_to_print, color=color, print_floors=print_floors)

def errprint(colored_string: str, stuff_to_print = '', color='RED', print_floors=False) -> None:
    qprint(colored_string=colored_string , stuff_to_print = stuff_to_print, color=color, print_floors=print_floors)

def qprint_array(arrray_to_print, rounding_decimal:int=3):
    """
    This function prints array in a human-readable format, removing artifacts such as "-0.", "0j" etc.
    Matrix is rounded up to rounding_decimal significant digits

    Parameters
    ----------
    arrray_to_print: npt.NDArray
    rounding_decimal: int

    Returns
    -------
    npt.NDArray
    """

    pd.set_option('display.max_columns', None)
    pd.set_option('display.width', 110)
    try:
        if (arrray_to_print.shape[0] == 1 or arrray_to_print.shape[1] == 1):
            B = copy.deepcopy(arrray_to_print)
            if (arrray_to_print.shape[0] == 1 and arrray_to_print.shape[1] == 1):
                print(np.round(arrray_to_print[0, 0], rounding_decimal))
            elif (arrray_to_print.shape[0] == 1):
                print([np.round(x[1], rounding_decimal) for x in arrray_to_print])
            elif (arrray_to_print.shape[1] == 1):
                print([np.round(x[0], rounding_decimal) for x in arrray_to_print])
        else:
            B = copy.deepcopy(arrray_to_print)
            C = round_matrix_for_print(B, rounding_decimal)
            D = zeros_to_dots_for_print(C, rounding_decimal)
            print(pd.DataFrame(D))
    except(IndexError):
        if len(arrray_to_print.shape) == 1:
            print([np.round(x, rounding_decimal) for x in arrray_to_print])
        else:
            print(pd.DataFrame(np.array(np.round(arrray_to_print, rounding_decimal))))


